#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from glanceclient import exc
from heat.common import exception
from heat.common import template_format
from heat.engine import stack as parser
from heat.engine import template
from heat.tests import common
from heat.tests import utils

image_download_template = '''
heat_template_version: rocky
description: This template to define a glance image.
resources:
  my_image:
    type: OS::Glance::WebImage
    properties:
      name: cirros_image
      id: 41f0e60c-ebb4-4375-a2b4-845ae8b9c995
      disk_format: qcow2
      container_format: bare
      min_disk: 10
      min_ram: 512
      protected: False
      location: https://launchpad.net/cirros/cirros-0.3.0-x86_64-disk.img
      architecture: test_architecture
      kernel_id: 12345678-1234-1234-1234-123456789012
      os_distro: test_distro
      owner: test_owner
      ramdisk_id: 12345678-1234-1234-1234-123456789012
'''

image_download_template_validate = '''
heat_template_version: rocky
description: This template to define a glance image.
resources:
  image:
    type: OS::Glance::WebImage
    properties:
      name: image_validate
      disk_format: qcow2
      container_format: bare
      location: https://launchpad.net/cirros/cirros-0.3.0-x86_64-disk.img
'''


class GlanceWebImageTest(common.HeatTestCase):
    def setUp(self):
        super(GlanceWebImageTest, self).setUp()

        self.ctx = utils.dummy_context()
        tpl = template_format.parse(image_download_template)
        self.stack = parser.Stack(
            self.ctx, 'glance_image_test_stack',
            template.Template(tpl)
        )

        self.my_image = self.stack['my_image']
        glance = mock.MagicMock()
        self.glanceclient = mock.MagicMock()
        self.my_image.client = glance
        glance.return_value = self.glanceclient
        self.images = self.glanceclient.images
        self.image_tags = self.glanceclient.image_tags
        self.image_members = self.glanceclient.image_members
        self.update = self.glanceclient.update

    def _test_validate(self, resource, error_msg):
        exc = self.assertRaises(exception.StackValidationFailed,
                                resource.validate)
        self.assertIn(error_msg, str(exc))

    def test_invalid_min_disk(self):
        # invalid 'min_disk'
        tpl = template_format.parse(image_download_template_validate)
        stack = parser.Stack(
            self.ctx, 'glance_image_stack_validate',
            template.Template(tpl)
        )
        image = stack['image']
        props = stack.t.t['resources']['image']['properties'].copy()
        props['min_disk'] = -1
        image.t = image.t.freeze(properties=props)
        image.reparse()
        error_msg = ('Property error: resources.image.properties.min_disk: '
                     '-1 is out of range (min: 0, max: None)')
        self._test_validate(image, error_msg)

    def test_invalid_min_ram(self):
        # invalid 'min_ram'
        tpl = template_format.parse(image_download_template_validate)
        stack = parser.Stack(
            self.ctx, 'glance_image_stack_validate',
            template.Template(tpl)
        )
        image = stack['image']
        props = stack.t.t['resources']['image']['properties'].copy()
        props['min_ram'] = -1
        image.t = image.t.freeze(properties=props)
        image.reparse()
        error_msg = ('Property error: resources.image.properties.min_ram: '
                     '-1 is out of range (min: 0, max: None)')
        self._test_validate(image, error_msg)

    def test_miss_disk_format(self):
        # miss disk_format
        tpl = template_format.parse(image_download_template_validate)
        stack = parser.Stack(
            self.ctx, 'glance_image_stack_validate',
            template.Template(tpl)
        )
        image = stack['image']
        props = stack.t.t['resources']['image']['properties'].copy()
        del props['disk_format']
        image.t = image.t.freeze(properties=props)
        image.reparse()
        error_msg = 'Property disk_format not assigned'
        self._test_validate(image, error_msg)

    def test_invalid_disk_format(self):
        # invalid disk_format
        tpl = template_format.parse(image_download_template_validate)
        stack = parser.Stack(
            self.ctx, 'glance_image_stack_validate',
            template.Template(tpl)
        )
        image = stack['image']
        props = stack.t.t['resources']['image']['properties'].copy()
        props['disk_format'] = 'incorrect_format'
        image.t = image.t.freeze(properties=props)
        image.reparse()
        error_msg = ('Property error: '
                     'resources.image.properties.disk_format: '
                     '"incorrect_format" is not an allowed value')
        self._test_validate(image, error_msg)

    def test_miss_container_format(self):
        # miss container_format
        tpl = template_format.parse(image_download_template_validate)
        stack = parser.Stack(
            self.ctx, 'glance_image_stack_validate',
            template.Template(tpl)
        )
        image = stack['image']
        props = stack.t.t['resources']['image']['properties'].copy()
        del props['container_format']
        image.t = image.t.freeze(properties=props)
        image.reparse()
        error_msg = 'Property container_format not assigned'
        self._test_validate(image, error_msg)

    def test_invalid_container_format(self):
        # invalid container_format
        tpl = template_format.parse(image_download_template_validate)
        stack = parser.Stack(
            self.ctx, 'glance_image_stack_validate',
            template.Template(tpl)
        )
        image = stack['image']
        props = stack.t.t['resources']['image']['properties'].copy()
        props['container_format'] = 'incorrect_format'
        image.t = image.t.freeze(properties=props)
        image.reparse()
        error_msg = ('Property error: '
                     'resources.image.properties.container_format: '
                     '"incorrect_format" is not an allowed value')
        self._test_validate(image, error_msg)

    def test_miss_location(self):
        # miss location
        tpl = template_format.parse(image_download_template_validate)
        stack = parser.Stack(
            self.ctx, 'glance_image_stack_validate',
            template.Template(tpl)
        )
        image = stack['image']
        props = stack.t.t['resources']['image']['properties'].copy()
        del props['location']
        image.t = image.t.freeze(properties=props)
        image.reparse()
        error_msg = 'Property location not assigned'
        self._test_validate(image, error_msg)

    def test_invalid_disk_container_mix(self):
        tpl = template_format.parse(image_download_template_validate)
        stack = parser.Stack(
            self.ctx, 'glance_image_stack_validate',
            template.Template(tpl)
        )
        image = stack['image']
        props = stack.t.t['resources']['image']['properties'].copy()
        props['disk_format'] = 'raw'
        props['container_format'] = 'ari'
        image.t = image.t.freeze(properties=props)
        image.reparse()
        error_msg = ("Invalid mix of disk and container formats. When "
                     "setting a disk or container format to one of 'aki', "
                     "'ari', or 'ami', the container and disk formats must "
                     "match.")
        self._test_validate(image, error_msg)

    def test_image_handle_create(self):
        value = mock.MagicMock()
        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
        value.id = image_id
        self.images.create.return_value = value
        self.image_tags.update.return_value = None
        props = self.stack.t.t['resources']['my_image']['properties'].copy()
        props['tags'] = ['tag1']
        props['extra_properties'] = {"hw_firmware_type": "uefi"}
        self.my_image.t = self.my_image.t.freeze(properties=props)
        self.my_image.reparse()
        self.my_image.handle_create()

        self.assertEqual(image_id, self.my_image.resource_id)
        # assert that no tags pass when image create
        self.images.create.assert_called_once_with(
            architecture='test_architecture',
            container_format='bare',
            disk_format='qcow2',
            id='41f0e60c-ebb4-4375-a2b4-845ae8b9c995',
            kernel_id='12345678-1234-1234-1234-123456789012',
            os_distro='test_distro',
            ramdisk_id='12345678-1234-1234-1234-123456789012',
            visibility='private',
            min_disk=10,
            min_ram=512,
            name='cirros_image',
            protected=False,
            owner='test_owner',
            tags=['tag1']
        )
        self.images.update.assert_called_once_with(
            image_id, hw_firmware_type='uefi')

    def test_image_active_property_image_not_active(self):
        self.images.reactivate.return_value = None
        self.images.deactivate.return_value = None
        value = mock.MagicMock()
        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
        value.id = image_id
        value.status = 'pending'
        self.images.create.return_value = value
        self.my_image.handle_create()
        self.my_image.check_create_complete(image_id)
        self.images.deactivate.assert_not_called()

    def test_image_active_property_image_active_to_deactivate(self):
        self.images.reactivate.return_value = None
        self.images.deactivate.return_value = None
        value = mock.MagicMock()
        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
        value.id = image_id
        value.status = 'active'
        self.my_image.resource_id = image_id
        self.images.create.return_value = value
        self.images.get.return_value = value
        self.my_image.check_create_complete(False)
        self.images.deactivate.assert_called_once_with(
            self.my_image.resource_id)

    def test_image_active_property_image_status_killed(self):
        self.images.reactivate.return_value = None
        self.images.deactivate.return_value = None
        value = mock.MagicMock()
        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
        value.id = image_id
        value.status = 'killed'
        self.my_image.resource_id = image_id
        self.images.create.return_value = value
        self.images.get.return_value = value
        ex = self.assertRaises(exception.ResourceInError,
                               self.my_image.check_create_complete, False)
        self.assertIn('killed', ex.message)

    def _handle_update_image_props(self, prop_diff):
        self.my_image.handle_update(json_snippet=None,
                                    tmpl_diff=None,
                                    prop_diff=prop_diff)
        self.images.update.assert_called_once_with(
            self.my_image.resource_id,
            ['hw_firmware_type'],
            os_secure_boot='required'
        )

    def _handle_update_tags(self, prop_diff):
        self.my_image.handle_update(json_snippet=None,
                                    tmpl_diff=None,
                                    prop_diff=prop_diff)

        self.image_tags.update.assert_called_once_with(
            self.my_image.resource_id,
            'tag2'
        )
        self.image_tags.delete.assert_called_once_with(
            self.my_image.resource_id,
            'tag1'
        )

    def test_image_handle_update(self):
        self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        prop_diff = {
            'architecture': 'test_architecture',
            'kernel_id': '12345678-1234-1234-1234-123456789012',
            'os_distro': 'test_distro',
            'owner': 'test_owner',
            'ramdisk_id': '12345678-1234-1234-1234-123456789012'}

        self.my_image.handle_update(json_snippet=None,
                                    tmpl_diff=None,
                                    prop_diff=prop_diff)
        self.images.update.assert_called_with(
            self.my_image.resource_id,
            architecture='test_architecture',
            kernel_id='12345678-1234-1234-1234-123456789012',
            os_distro='test_distro',
            owner='test_owner',
            ramdisk_id='12345678-1234-1234-1234-123456789012'
        )

    def test_image_handle_update_deactivate(self):
        self.images.reactivate.return_value = None
        self.images.deactivate.return_value = None
        value = mock.MagicMock()
        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
        value.id = image_id
        value.status = 'active'
        self.my_image.resource_id = image_id
        props = self.stack.t.t['resources']['my_image']['properties'].copy()
        props['active'] = False
        self.my_image.t = self.my_image.t.freeze(properties=props)
        prop_diff = {'active': False}
        self.my_image.reparse()
        self.images.update.return_value = value
        self.images.get.return_value = value
        self.my_image.handle_update(json_snippet=None,
                                    tmpl_diff=None,
                                    prop_diff=prop_diff)
        self.images.deactivate.assert_called_once_with(
            self.my_image.resource_id)

    def test_image_handle_update_reactivate(self):
        self.images.reactivate.return_value = None
        self.images.deactivate.return_value = None
        value = mock.MagicMock()
        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
        value.id = image_id
        value.status = 'deactivated'
        self.my_image.resource_id = image_id
        props = self.stack.t.t['resources']['my_image']['properties'].copy()
        props['active'] = True
        self.my_image.t = self.my_image.t.freeze(properties=props)
        prop_diff = {'active': True}
        self.my_image.reparse()
        self.images.update.return_value = value
        self.images.get.return_value = value
        self.my_image.handle_update(json_snippet=None,
                                    tmpl_diff=None,
                                    prop_diff=prop_diff)
        self.my_image.check_update_complete(True)
        self.images.reactivate.assert_called_once_with(
            self.my_image.resource_id)

    def test_image_handle_update_image_props(self):
        self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        props = self.stack.t.t['resources']['my_image']['properties'].copy()
        props['extra_properties'] = {"hw_firmware_type": "uefi"}
        self.my_image.t = self.my_image.t.freeze(properties=props)
        self.my_image.reparse()
        prop_diff = {'extra_properties': {"os_secure_boot": "required"}}

        self._handle_update_image_props(prop_diff)

    def test_image_handle_update_tags(self):
        self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        props = self.stack.t.t['resources']['my_image']['properties'].copy()
        props['tags'] = ['tag1']
        self.my_image.t = self.my_image.t.freeze(properties=props)
        self.my_image.reparse()
        prop_diff = {'tags': ['tag2']}

        self._handle_update_tags(prop_diff)

    def test_image_handle_update_remove_tags(self):
        self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        props = self.stack.t.t['resources']['my_image']['properties'].copy()
        props['tags'] = ['tag1']
        self.my_image.t = self.my_image.t.freeze(properties=props)
        self.my_image.reparse()
        prop_diff = {'tags': None}

        self.my_image.handle_update(json_snippet=None,
                                    tmpl_diff=None,
                                    prop_diff=prop_diff)

        self.image_tags.delete.assert_called_once_with(
            self.my_image.resource_id,
            'tag1'
        )

    def _handle_update_members(self, prop_diff):
        self.my_image.handle_update(json_snippet=None,
                                    tmpl_diff=None,
                                    prop_diff=prop_diff)

        self.image_members.create.assert_called_once_with(
            self.my_image.resource_id,
            'member2'
        )
        self.image_members.delete.assert_called_once_with(
            self.my_image.resource_id,
            'member1'
        )

    def test_image_handle_update_members(self):
        self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        props = self.stack.t.t['resources']['my_image']['properties'].copy()
        props['members'] = ['member1']
        self.my_image.t = self.my_image.t.freeze(properties=props)
        self.my_image.reparse()
        prop_diff = {'members': ['member2']}

        self._handle_update_members(prop_diff)

    def test_image_handle_update_remove_members(self):
        self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        props = self.stack.t.t['resources']['my_image']['properties'].copy()
        props['members'] = ['member1']
        self.my_image.t = self.my_image.t.freeze(properties=props)
        self.my_image.reparse()
        prop_diff = {'members': None}

        self.my_image.handle_update(json_snippet=None,
                                    tmpl_diff=None,
                                    prop_diff=prop_diff)

        self.image_members.delete.assert_called_once_with(
            self.my_image.resource_id,
            'member1'
        )

    def test_image_handle_update_tags_delete_not_found(self):
        self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        props = self.stack.t.t['resources']['my_image']['properties'].copy()
        props['tags'] = ['tag1']
        self.my_image.t = self.my_image.t.freeze(properties=props)
        self.my_image.reparse()
        prop_diff = {'tags': ['tag2']}

        self.image_tags.delete.side_effect = exc.HTTPNotFound()

        self._handle_update_tags(prop_diff)

    def test_image_show_resource_v1(self):
        self.glanceclient.version = 1.0
        self.my_image.resource_id = 'test_image_id'
        image = mock.MagicMock()
        images = mock.MagicMock()
        image.to_dict.return_value = {'image': 'info'}
        images.get.return_value = image
        self.my_image.client().images = images
        self.assertEqual({'image': 'info'}, self.my_image.FnGetAtt('show'))
        images.get.assert_called_once_with('test_image_id')

    def test_image_show_resource_v2(self):
        self.my_image.resource_id = 'test_image_id'
        # glance image in v2 is warlock.model object, so it can be
        # handled via dict(). In test we use easiest analog - dict.
        image = {"key1": "val1", "key2": "val2"}
        self.images.get.return_value = image
        self.glanceclient.version = 2.0
        self.assertEqual({"key1": "val1", "key2": "val2"},
                         self.my_image.FnGetAtt('show'))
        self.images.get.assert_called_once_with('test_image_id')

    def test_image_get_live_state_v2(self):
        self.glanceclient.version = 2.0
        self.my_image.resource_id = '1234'
        images = mock.MagicMock()
        show_value = {
            'name': 'test',
            'disk_format': 'qcow2',
            'container_format': 'bare',
            'active': None,
            'protected': False,
            'is_public': False,
            'min_disk': 0,
            'min_ram': 0,
            'id': '41f0e60c-ebb4-4375-a2b4-845ae8b9c995',
            'tags': [],
            'architecture': 'test_architecture',
            'kernel_id': '12345678-1234-1234-1234-123456789012',
            'os_distro': 'test_distro',
            'os_version': '1.0',
            'owner': 'test_owner',
            'ramdisk_id': '12345678-1234-1234-1234-123456789012',
            'members': None,
            'visibility': 'private'
        }
        image = show_value
        images.get.return_value = image
        self.my_image.client().images = images

        reality = self.my_image.get_live_state(self.my_image.properties)
        expected = {
            'name': 'test',
            'disk_format': 'qcow2',
            'container_format': 'bare',
            'active': None,
            'protected': False,
            'min_disk': 0,
            'min_ram': 0,
            'id': '41f0e60c-ebb4-4375-a2b4-845ae8b9c995',
            'tags': [],
            'architecture': 'test_architecture',
            'kernel_id': '12345678-1234-1234-1234-123456789012',
            'os_distro': 'test_distro',
            'os_version': '1.0',
            'owner': 'test_owner',
            'ramdisk_id': '12345678-1234-1234-1234-123456789012',
            'members': None,
            'visibility': 'private'
        }

        self.assertEqual(set(expected.keys()), set(reality.keys()))
        for key in expected:
            self.assertEqual(expected[key], reality[key])

    def test_get_live_state_resource_is_deleted(self):
        self.my_image.resource_id = '1234'
        self.my_image.client().images.get.return_value = {'status': 'deleted'}
        self.assertRaises(exception.EntityNotFound,
                          self.my_image.get_live_state,
                          self.my_image.properties)

    def test_parse_live_resource_data(self):
        resource_data = {
            'name': 'test',
            'disk_format': 'qcow2',
            'container_format': 'bare',
            'active': None,
            'protected': False,
            'is_public': False,
            'min_disk': 0,
            'min_ram': 0,
            'id': '41f0e60c-ebb4-4375-a2b4-845ae8b9c995',
            'tags': [],
            'architecture': 'test_architecture',
            'kernel_id': '12345678-1234-1234-1234-123456789012',
            'os_distro': 'new_distro',
            'os_version': '1.0',
            'os_secure_boot': 'False',
            'owner': 'new_owner',
            'hw_firmware_type': 'uefi',
            'ramdisk_id': '12345678-1234-1234-1234-123456789012',
            'members': None,
            'visibility': 'private'
        }

        resource_properties = self.stack.t.t['resources'][
            'my_image']['properties'].copy()
        resource_properties['extra_properties'] = {
            'hw_firmware_type': 'uefi',
            'os_secure_boot': 'required',
            }

        reality = self.my_image.parse_live_resource_data(resource_properties,
                                                         resource_data)
        expected = {
            'name': 'test',
            'disk_format': 'qcow2',
            'container_format': 'bare',
            'active': None,
            'protected': False,
            'min_disk': 0,
            'min_ram': 0,
            'id': '41f0e60c-ebb4-4375-a2b4-845ae8b9c995',
            'tags': [],
            'architecture': 'test_architecture',
            'kernel_id': '12345678-1234-1234-1234-123456789012',
            'os_distro': 'new_distro',
            'os_version': '1.0',
            'owner': 'new_owner',
            'ramdisk_id': '12345678-1234-1234-1234-123456789012',
            'members': None,
            'visibility': 'private',
            'extra_properties': {
                'hw_firmware_type': 'uefi',
                'os_secure_boot': 'False',
            }
        }

        self.assertEqual(set(expected.keys()), set(reality.keys()))
        for key in expected:
            self.assertEqual(expected[key], reality[key])
        for key in expected['extra_properties']:
            self.assertEqual(expected['extra_properties'][key],
                             reality['extra_properties'][key])



# part-handler
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import errno
import os
import sys


def list_types():
    return ["text/x-cfninitdata"]


def handle_part(data, ctype, filename, payload):
    if ctype == "__begin__":
        try:
            os.makedirs('/var/lib/heat-cfntools', int("700", 8))
        except OSError:
            ex_type, e, tb = sys.exc_info()
            if e.errno != errno.EEXIST:
                raise
        return

    if ctype == "__end__":
        return

    timestamp = datetime.datetime.now()
    with open('/var/log/part-handler.log', 'a') as log:
        log.write('%s filename:%s, ctype:%s\n' % (timestamp, filename, ctype))

    if ctype == 'text/x-cfninitdata':
        with open('/var/lib/heat-cfntools/%s' % filename, 'w') as f:
            f.write(payload)

        # TODO(sdake) hopefully temporary until users move to heat-cfntools-1.3
        with open('/var/lib/cloud/data/%s' % filename, 'w') as f:
            f.write(payload)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from urllib import parse

from keystoneclient.contrib.ec2 import utils as ec2_utils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import timeutils

from heat.common import exception
from heat.common.i18n import _
from heat.common import password_gen
from heat.engine.clients.os import swift
from heat.engine.resources import stack_user

LOG = logging.getLogger(__name__)

SIGNAL_TYPES = (
    WAITCONDITION, SIGNAL
) = (
    '/waitcondition', '/signal'
)
SIGNAL_VERB = {WAITCONDITION: 'PUT',
               SIGNAL: 'POST'}


class SignalResponder(stack_user.StackUser):

    PROPERTIES = (
        SIGNAL_TRANSPORT,
    ) = (
        'signal_transport',
    )

    ATTRIBUTES = (
        SIGNAL_ATTR,
    ) = (
        'signal',
    )

    # Anything which subclasses this may trigger authenticated
    # API operations as a consequence of handling a signal
    requires_deferred_auth = True

    def handle_delete(self):
        self._delete_signals()
        return super(SignalResponder, self).handle_delete()

    def _delete_signals(self):
        self._delete_ec2_signed_url()
        self._delete_heat_signal_url()
        self._delete_swift_signal_url()
        self._delete_zaqar_signal_queue()

    @property
    def password(self):
        return self.data().get('password')

    @password.setter
    def password(self, password):
        if password is None:
            self.data_delete('password')
        else:
            self.data_set('password', password, True)

    def _signal_transport_cfn(self):
        return self.properties[
            self.SIGNAL_TRANSPORT] == self.CFN_SIGNAL

    def _signal_transport_heat(self):
        return self.properties[
            self.SIGNAL_TRANSPORT] == self.HEAT_SIGNAL

    def _signal_transport_none(self):
        return self.properties[
            self.SIGNAL_TRANSPORT] == self.NO_SIGNAL

    def _signal_transport_temp_url(self):
        return self.properties[
            self.SIGNAL_TRANSPORT] == self.TEMP_URL_SIGNAL

    def _signal_transport_zaqar(self):
        return self.properties.get(
            self.SIGNAL_TRANSPORT) == self.ZAQAR_SIGNAL

    def _get_region_name(self):
        return self.client_plugin('heat')._get_region_name()

    def _get_heat_signal_credentials(self):
        """Return OpenStack credentials that can be used to send a signal.

        These credentials are for the user associated with this resource in
        the heat stack user domain.
        """
        if self._get_user_id() is None:
            if self.password is None:
                self.password = password_gen.generate_openstack_password()
            self._create_user()
        return {'auth_url': self.keystone().server_keystone_endpoint_url(
                fallback_endpoint=self.keystone().v3_endpoint),
                'username': self.physical_resource_name(),
                'user_id': self._get_user_id(),
                'password': self.password,
                'project_id': self.stack.stack_user_project_id,
                'domain_id': self.keystone().stack_domain_id,
                'region_name': self._get_region_name()}

    def _get_ec2_signed_url(self, signal_type=SIGNAL,
                            never_expire=False):
        """Create properly formatted and pre-signed URL.

        This uses the created user for the credentials.

        See boto/auth.py::QuerySignatureV2AuthHandler

        :param signal_type: either WAITCONDITION or SIGNAL.
        """
        access_key = self.data().get('access_key')
        secret_key = self.data().get('secret_key')

        if not access_key or not secret_key:
            if self.id is None:
                # it is too early
                return
            if self._get_user_id() is None:
                self._create_user()
            self._create_keypair()
            access_key = self.data().get('access_key')
            secret_key = self.data().get('secret_key')

            if not access_key or not secret_key:
                LOG.warning('Cannot generate signed url, '
                            'unable to create keypair')
                return

        config_url = cfg.CONF.heat_waitcondition_server_url
        if config_url:
            signal_url = config_url.replace('/waitcondition', signal_type)
        else:
            heat_client_plugin = self.stack.clients.client_plugin('heat')
            endpoint = heat_client_plugin.get_heat_cfn_url()
            signal_url = ''.join([endpoint, signal_type])

        host_url = parse.urlparse(signal_url)

        path = self.identifier().arn_url_path()

        # Note the WSGI spec apparently means that the webob request we end up
        # processing in the CFN API (ec2token.py) has an unquoted path, so we
        # need to calculate the signature with the path component unquoted, but
        # ensure the actual URL contains the quoted version...
        unquoted_path = parse.unquote(host_url.path + path)
        params = {'SignatureMethod': 'HmacSHA256',
                  'SignatureVersion': '2',
                  'AWSAccessKeyId': access_key}
        if not never_expire:
            params['Timestamp'] = timeutils.utcnow().strftime(
                "%Y-%m-%dT%H:%M:%SZ")
        request = {'host': host_url.netloc.lower(),
                   'verb': SIGNAL_VERB[signal_type],
                   'path': unquoted_path,
                   'params': params}
        # Sign the request
        signer = ec2_utils.Ec2Signer(secret_key)
        request['params']['Signature'] = signer.generate(request)

        qs = parse.urlencode(request['params'])
        url = "%s%s?%s" % (signal_url.lower(),
                           path, qs)

        return url

    def _delete_ec2_signed_url(self):
        self.data_delete('ec2_signed_url')
        self._delete_keypair()

    def _get_heat_signal_url(self, project_id=None):
        """Return a heat-api signal URL for this resource.

        This URL is not pre-signed, valid user credentials are required.
        If a project_id is provided, it is used in place of the original
        project_id. This is useful to generate a signal URL that uses
        the heat stack user project instead of the user's.
        """
        stored = self.data().get('heat_signal_url')
        if stored is not None:
            return stored

        if self.id is None:
            # it is too early
            return

        url = self.client_plugin('heat').get_heat_url()
        path = self.identifier().url_path()
        if project_id is not None:
            path = project_id + path[path.find('/'):]

        url = parse.urljoin(url, '%s/signal' % path)

        self.data_set('heat_signal_url', url)
        return url

    def _delete_heat_signal_url(self):
        self.data_delete('heat_signal_url')

    def _get_swift_signal_url(self, multiple_signals=False):
        """Create properly formatted and pre-signed Swift signal URL.

        This uses a Swift pre-signed temp_url. If multiple_signals is
        requested, the Swift object referenced by the returned URL will have
        versioning enabled.
        """
        put_url = self.data().get('swift_signal_url')
        if put_url:
            return put_url

        if self.id is None:
            # it is too early
            return

        container = self.stack.id
        object_name = self.physical_resource_name()

        self.client('swift').put_container(container)

        if multiple_signals:
            put_url = self.client_plugin('swift').get_signal_url(container,
                                                                 object_name)
        else:
            put_url = self.client_plugin('swift').get_temp_url(container,
                                                               object_name)
            self.client('swift').put_object(container, object_name, '')
        self.data_set('swift_signal_url', put_url)
        self.data_set('swift_signal_object_name', object_name)

        return put_url

    def _delete_swift_signal_url(self):
        object_name = self.data().get('swift_signal_object_name')
        if not object_name:
            return
        with self.client_plugin('swift').ignore_not_found:
            container_name = self.stack.id
            swift = self.client('swift')

            # delete all versions of the object, in case there are some
            # signals that are waiting to be handled
            container = swift.get_container(container_name)
            filtered = [obj for obj in container[1]
                        if object_name in obj['name']]
            for obj in filtered:
                # we delete the main object every time, swift takes
                # care of restoring the previous version after each delete
                swift.delete_object(container_name, object_name)

            headers = swift.head_container(container_name)
            if int(headers['x-container-object-count']) == 0:
                swift.delete_container(container_name)
        self.data_delete('swift_signal_object_name')
        self.data_delete('swift_signal_url')

    def _get_zaqar_signal_queue_id(self):
        """Return a zaqar queue_id for signaling this resource.

        This uses the created user for the credentials.
        """
        queue_id = self.data().get('zaqar_signal_queue_id')
        if queue_id:
            return queue_id

        if self.id is None:
            # it is too early
            return

        if self._get_user_id() is None:
            if self.password is None:
                self.password = password_gen.generate_openstack_password()
            self._create_user()

        queue_id = self.physical_resource_name()
        zaqar_plugin = self.client_plugin('zaqar')
        zaqar = zaqar_plugin.create_for_tenant(
            self.stack.stack_user_project_id, self._user_token())
        queue = zaqar.queue(queue_id)
        signed_url_data = queue.signed_url(
            ['messages'], methods=['GET', 'DELETE'])
        self.data_set('zaqar_queue_signed_url_data',
                      jsonutils.dumps(signed_url_data))
        self.data_set('zaqar_signal_queue_id', queue_id)
        return queue_id

    def _delete_zaqar_signal_queue(self):
        queue_id = self.data().get('zaqar_signal_queue_id')
        if not queue_id:
            return
        zaqar_plugin = self.client_plugin('zaqar')
        zaqar = zaqar_plugin.create_for_tenant(
            self.stack.stack_user_project_id, self._user_token())
        with zaqar_plugin.ignore_not_found:
            zaqar.queue(queue_id).delete()
        self.data_delete('zaqar_signal_queue_id')

    def _get_signal(self, signal_type=SIGNAL, multiple_signals=False):
        """Return a dictionary with signal details.

        Subclasses can invoke this method to retrieve information of the
        resource signal for the specified transport.
        """
        signal = None
        if self._signal_transport_cfn():
            signal = {'alarm_url': self._get_ec2_signed_url(
                signal_type=signal_type)}
        elif self._signal_transport_heat():
            signal = self._get_heat_signal_credentials()
            signal['alarm_url'] = self._get_heat_signal_url(
                project_id=self.stack.stack_user_project_id)
        elif self._signal_transport_temp_url():
            signal = {'alarm_url': self._get_swift_signal_url(
                multiple_signals=multiple_signals)}
        elif self._signal_transport_zaqar():
            signal = self._get_heat_signal_credentials()
            signal['queue_id'] = self._get_zaqar_signal_queue_id()
        elif self._signal_transport_none():
            signal = {}
        return signal

    def _service_swift_signal(self):
        swift_client = self.client('swift')
        try:
            container = swift_client.get_container(self.stack.id)
        except Exception as exc:
            self.client_plugin('swift').ignore_not_found(exc)
            LOG.debug("Swift container %s was not found", self.stack.id)
            return

        index = container[1]
        if not index:  # Swift objects were deleted by user
            LOG.debug("Swift objects in container %s were not found",
                      self.stack.id)
            return

        # Remove objects that are for other resources, given that
        # multiple swift signals in the same stack share a container
        object_name = self.physical_resource_name()
        filtered = [obj for obj in index if object_name in obj['name']]

        # Fetch objects from Swift and filter results
        signal_names = []
        for obj in filtered:
            try:
                signal = swift_client.get_object(self.stack.id, obj['name'])
            except Exception as exc:
                self.client_plugin('swift').ignore_not_found(exc)
                continue

            body = signal[1]
            if body == swift.IN_PROGRESS:  # Ignore the initial object
                continue
            signal_names.append(obj['name'])

            if body == "":
                self.signal(details={})
                continue
            try:
                self.signal(details=jsonutils.loads(body))
            except ValueError:
                raise exception.Error(_("Failed to parse JSON data: %s") %
                                      body)

        # remove the signals that were consumed
        for signal_name in signal_names:
            if signal_name != object_name:
                swift_client.delete_object(self.stack.id, signal_name)
        if object_name in signal_names:
            swift_client.delete_object(self.stack.id, object_name)

    def _service_zaqar_signal(self):
        zaqar_plugin = self.client_plugin('zaqar')
        zaqar = zaqar_plugin.create_for_tenant(
            self.stack.stack_user_project_id, self._user_token())
        try:
            queue = zaqar.queue(self._get_zaqar_signal_queue_id())
        except Exception as ex:
            self.client_plugin('zaqar').ignore_not_found(ex)
            return
        messages = list(queue.pop())
        for message in messages:
            self.signal(details=message.body)

    def _service_signal(self):
        """Service the signal, when necessary.

        This method must be called repeatedly by subclasses to update the
        state of the signals that require polling, which are the ones based on
        Swift temp URLs and Zaqar queues. The "NO_SIGNAL" case is also handled
        here by triggering the signal once per call.
        """
        if self._signal_transport_temp_url():
            self._service_swift_signal()
        elif self._signal_transport_zaqar():
            self._service_zaqar_signal()
        elif self._signal_transport_none():
            self.signal(details={})



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import swiftclient.client as sc

from heat.common import exception
from heat.common import template_format
from heat.engine import node_data
from heat.engine.resources.openstack.swift import container as swift_c
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


SWIFT_TEMPLATE = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "Template to test OS::Swift::Container resources",
  "Resources" : {
    "SwiftContainerWebsite" : {
      "Type" : "OS::Swift::Container",
      "DeletionPolicy" : "Delete",
      "Properties" : {
        "X-Container-Read" : ".r:*",
        "X-Container-Meta" : {
          "Web-Index" : "index.html",
          "Web-Error" : "error.html"
}
      }
    },
    "SwiftAccountMetadata" : {
      "Type" : "OS::Swift::Container",
      "DeletionPolicy" : "Delete",
      "Properties" : {
        "X-Account-Meta" : {
          "Temp-Url-Key" : "secret"
         }
      }
    },
    "S3Bucket" : {
      "Type" : "AWS::S3::Bucket",
      "Properties" : {
        "SwiftContainer" : {"Ref" : "SwiftContainer"}
      }
    },
    "SwiftContainer" : {
      "Type" : "OS::Swift::Container",
      "Properties" : {
      }
    }
  }
}
'''


class SwiftTest(common.HeatTestCase):

    def setUp(self):
        super(SwiftTest, self).setUp()
        self.t = template_format.parse(SWIFT_TEMPLATE)

    def _create_container(self, stack, definition_name='SwiftContainer'):
        resource_defns = stack.t.resource_definitions(stack)
        container = swift_c.SwiftContainer('test_resource',
                                           resource_defns[definition_name],
                                           stack)
        runner = scheduler.TaskRunner(container.create)
        runner()
        self.assertEqual((container.CREATE, container.COMPLETE),
                         container.state)
        return container

    @mock.patch('swiftclient.client.Connection.put_container')
    def test_create_container_name(self, mock_put):
        # Setup
        res_prop = self.t['Resources']['SwiftContainer']['Properties']
        res_prop['name'] = 'the_name'
        stack = utils.parse_stack(self.t)

        # Test
        container = self._create_container(stack)
        container_name = container.physical_resource_name()

        # Verify
        self.assertEqual('the_name', container_name)
        mock_put.assert_called_once_with('the_name', {})

    def test_build_meta_headers(self):
        # Setup
        headers = {'Web-Index': 'index.html', 'Web-Error': 'error.html'}

        # Test
        self.assertEqual({}, swift_c.SwiftContainer._build_meta_headers(
            'container', {}))
        self.assertEqual({}, swift_c.SwiftContainer._build_meta_headers(
            'container', None))
        built = swift_c.SwiftContainer._build_meta_headers(
            'container', headers)

        # Verify
        expected = {
            'X-Container-Meta-Web-Index': 'index.html',
            'X-Container-Meta-Web-Error': 'error.html'
        }
        self.assertEqual(expected, built)

    @mock.patch('swiftclient.client.Connection.head_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_attributes(self, mock_put, mock_head):
        # Setup
        headers = {'content-length': '0',
                   'x-container-object-count': '82',
                   'accept-ranges': 'bytes',
                   'x-trans-id': 'tx08ea48ef2fa24e6da3d2f5c188fd938b',
                   'date': 'Wed, 23 Jan 2013 22:48:05 GMT',
                   'x-timestamp': '1358980499.84298',
                   'x-container-read': '.r:*',
                   'x-container-bytes-used': '17680980',
                   'content-type': 'text/plain; charset=utf-8'}
        mock_head.return_value = headers

        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        # Test
        container = self._create_container(stack)

        # call this to populate the url of swiftclient. This is actually
        # set in head_container/put_container, but we're patching them in
        # this test.
        container.client().get_auth()

        # Verify Attributes
        self.assertEqual(container_name, container.FnGetRefId())
        self.assertEqual('82', container.FnGetAtt('ObjectCount'))
        self.assertEqual('17680980', container.FnGetAtt('BytesUsed'))
        self.assertEqual('server.test', container.FnGetAtt('DomainName'))
        self.assertEqual(headers, container.FnGetAtt('HeadContainer'))
        self.assertEqual(headers, container.FnGetAtt('show'))

        expected_url = 'http://server.test:5000/v3/%s' % container.FnGetRefId()
        self.assertEqual(expected_url, container.FnGetAtt('WebsiteURL'))

        self.assertRaises(exception.InvalidTemplateAttribute,
                          container.FnGetAtt, 'Foo')

        # Verify Expected Calls
        mock_put.assert_called_once_with(container_name, {})
        self.assertGreater(mock_head.call_count, 0)

    @mock.patch('swiftclient.client.Connection.put_container')
    def test_public_read(self, mock_put):
        # Setup
        properties = self.t['Resources']['SwiftContainer']['Properties']
        properties['X-Container-Read'] = '.r:*'
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        # Test
        self._create_container(stack)

        # Verify
        expected = {'X-Container-Read': '.r:*'}
        mock_put.assert_called_once_with(container_name, expected)

    @mock.patch('swiftclient.client.Connection.put_container')
    def test_public_read_write(self, mock_put):
        # Setup
        properties = self.t['Resources']['SwiftContainer']['Properties']
        properties['X-Container-Read'] = '.r:*'
        properties['X-Container-Write'] = '.r:*'
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        # Test
        self._create_container(stack)

        # Verify
        expected = {'X-Container-Write': '.r:*', 'X-Container-Read': '.r:*'}
        mock_put.assert_called_once_with(container_name, expected)

    @mock.patch('swiftclient.client.Connection.put_container')
    def test_container_headers(self, mock_put):
        # Setup
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        # Test
        self._create_container(stack,
                               definition_name='SwiftContainerWebsite')

        # Verify
        expected = {'X-Container-Meta-Web-Error': 'error.html',
                    'X-Container-Meta-Web-Index': 'index.html',
                    'X-Container-Read': '.r:*'}
        mock_put.assert_called_once_with(container_name, expected)

    @mock.patch('swiftclient.client.Connection.post_account')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_account_headers(self, mock_put, mock_post):
        # Setup
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        # Test
        self._create_container(stack,
                               definition_name='SwiftAccountMetadata')

        # Verify
        mock_put.assert_called_once_with(container_name, {})
        expected = {'X-Account-Meta-Temp-Url-Key': 'secret'}
        mock_post.assert_called_once_with(expected)

    @mock.patch('swiftclient.client.Connection.put_container')
    def test_default_headers_not_none_empty_string(self, mock_put):
        # Setup
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        # Test
        container = self._create_container(stack)

        # Verify
        mock_put.assert_called_once_with(container_name, {})
        self.assertEqual({}, container.metadata_get())

    @mock.patch('swiftclient.client.Connection.delete_container')
    @mock.patch('swiftclient.client.Connection.get_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_delete_exception(self, mock_put, mock_get, mock_delete):
        # Setup
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        mock_delete.side_effect = sc.ClientException('test-delete-failure')
        mock_get.return_value = ({'name': container_name}, [])

        # Test
        container = self._create_container(stack)
        runner = scheduler.TaskRunner(container.delete)
        self.assertRaises(exception.ResourceFailure, runner)

        # Verify
        self.assertEqual((container.DELETE, container.FAILED),
                         container.state)
        mock_put.assert_called_once_with(container_name, {})
        mock_get.assert_called_once_with(container_name)
        mock_delete.assert_called_once_with(container_name)

    @mock.patch('swiftclient.client.Connection.delete_container')
    @mock.patch('swiftclient.client.Connection.get_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_delete_not_found(self, mock_put, mock_get, mock_delete):
        # Setup
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        mock_delete.side_effect = sc.ClientException('missing',
                                                     http_status=404)
        mock_get.return_value = ({'name': container_name}, [])

        # Test
        container = self._create_container(stack)
        runner = scheduler.TaskRunner(container.delete)
        runner()

        # Verify
        self.assertEqual((container.DELETE, container.COMPLETE),
                         container.state)
        mock_put.assert_called_once_with(container_name, {})
        mock_get.assert_called_once_with(container_name)
        mock_delete.assert_called_once_with(container_name)

    @mock.patch('swiftclient.client.Connection.get_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_delete_non_empty_not_allowed(self, mock_put, mock_get):
        # Setup
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        mock_get.return_value = ({'name': container_name},
                                 [{'name': 'test_object'}])

        # Test
        container = self._create_container(stack)
        runner = scheduler.TaskRunner(container.delete)
        ex = self.assertRaises(exception.ResourceFailure, runner)

        # Verify
        self.assertEqual((container.DELETE, container.FAILED),
                         container.state)
        self.assertIn('ResourceActionNotSupported: resources.test_resource: '
                      'Deleting non-empty container',
                      str(ex))
        mock_put.assert_called_once_with(container_name, {})
        mock_get.assert_called_once_with(container_name)

    @mock.patch('swiftclient.client.Connection.delete_container')
    @mock.patch('swiftclient.client.Connection.delete_object')
    @mock.patch('swiftclient.client.Connection.get_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_delete_non_empty_allowed(self, mock_put, mock_get,
                                      mock_delete_object,
                                      mock_delete_container):
        # Setup
        res_prop = self.t['Resources']['SwiftContainer']['Properties']
        res_prop['PurgeOnDelete'] = True
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        get_return_values = [
            ({'name': container_name},
             [{'name': 'test_object1'},
              {'name': 'test_object2'}]),
            ({'name': container_name}, [{'name': 'test_object1'}]),
        ]
        mock_get.side_effect = get_return_values

        # Test
        container = self._create_container(stack)
        runner = scheduler.TaskRunner(container.delete)
        runner()

        # Verify
        self.assertEqual((container.DELETE, container.COMPLETE),
                         container.state)
        mock_put.assert_called_once_with(container_name, {})
        mock_delete_container.assert_called_once_with(container_name)
        self.assertEqual(2, mock_get.call_count)
        self.assertEqual(2, mock_delete_object.call_count)

    @mock.patch('swiftclient.client.Connection.delete_container')
    @mock.patch('swiftclient.client.Connection.delete_object')
    @mock.patch('swiftclient.client.Connection.get_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_delete_non_empty_allowed_not_found(self, mock_put, mock_get,
                                                mock_delete_object,
                                                mock_delete_container):
        # Setup
        res_prop = self.t['Resources']['SwiftContainer']['Properties']
        res_prop['PurgeOnDelete'] = True
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        mock_get.return_value = ({'name': container_name},
                                 [{'name': 'test_object'}])
        mock_delete_object.side_effect = sc.ClientException('object-is-gone',
                                                            http_status=404)
        mock_delete_container.side_effect = sc.ClientException(
            'container-is-gone', http_status=404)

        # Test
        container = self._create_container(stack)
        runner = scheduler.TaskRunner(container.delete)
        runner()

        # Verify
        self.assertEqual((container.DELETE, container.COMPLETE),
                         container.state)
        mock_put.assert_called_once_with(container_name, {})
        mock_get.assert_called_once_with(container_name)
        mock_delete_object.assert_called_once_with(container_name,
                                                   'test_object')
        mock_delete_container.assert_called_once_with(container_name)

    @mock.patch('swiftclient.client.Connection.delete_object')
    @mock.patch('swiftclient.client.Connection.get_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_delete_non_empty_fails_delete_object(self, mock_put, mock_get,
                                                  mock_delete_object):
        # Setup
        res_prop = self.t['Resources']['SwiftContainer']['Properties']
        res_prop['PurgeOnDelete'] = True
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        mock_get.return_value = ({'name': container_name},
                                 [{'name': 'test_object'}])
        mock_delete_object.side_effect = (
            sc.ClientException('object-delete-failure'))

        # Test
        container = self._create_container(stack)
        runner = scheduler.TaskRunner(container.delete)
        self.assertRaises(exception.ResourceFailure, runner)

        # Verify
        self.assertEqual((container.DELETE, container.FAILED),
                         container.state)
        mock_put.assert_called_once_with(container_name, {})
        mock_get.assert_called_once_with(container_name)
        mock_delete_object.assert_called_once_with(container_name,
                                                   'test_object')

    @mock.patch('swiftclient.client.Connection.put_container')
    def test_delete_retain(self, mock_put):
        # Setup
        self.t['Resources']['SwiftContainer']['DeletionPolicy'] = 'Retain'
        stack = utils.parse_stack(self.t)
        container_name = utils.PhysName(stack.name, 'test_resource')

        # Test
        container = self._create_container(stack)
        runner = scheduler.TaskRunner(container.delete)
        runner()

        # Verify
        self.assertEqual((container.DELETE, container.COMPLETE),
                         container.state)
        mock_put.assert_called_once_with(container_name, {})

    @mock.patch('swiftclient.client.Connection.get_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_check(self, mock_put, mock_get):
        # Setup
        res_prop = self.t['Resources']['SwiftContainer']['Properties']
        res_prop['PurgeOnDelete'] = True
        stack = utils.parse_stack(self.t)

        # Test
        container = self._create_container(stack)
        runner = scheduler.TaskRunner(container.check)
        runner()
        self.assertEqual((container.CHECK, container.COMPLETE),
                         container.state)

    @mock.patch('swiftclient.client.Connection.get_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_check_fail(self, mock_put, mock_get):
        # Setup
        res_prop = self.t['Resources']['SwiftContainer']['Properties']
        res_prop['PurgeOnDelete'] = True
        stack = utils.parse_stack(self.t)

        mock_get.side_effect = Exception('boom')

        # Test
        container = self._create_container(stack)
        runner = scheduler.TaskRunner(container.check)
        ex = self.assertRaises(exception.ResourceFailure, runner)

        # Verify
        self.assertIn('boom', str(ex))
        self.assertEqual((container.CHECK, container.FAILED),
                         container.state)

    def test_refid(self):
        stack = utils.parse_stack(self.t)
        rsrc = stack['SwiftContainer']
        rsrc.resource_id = 'xyz'
        self.assertEqual('xyz', rsrc.FnGetRefId())

    def test_refid_convergence_cache_data(self):
        cache_data = {'SwiftContainer': node_data.NodeData.from_dict({
            'uuid': mock.ANY,
            'id': mock.ANY,
            'action': 'CREATE',
            'status': 'COMPLETE',
            'reference_id': 'xyz_convg'
        })}
        stack = utils.parse_stack(self.t, cache_data=cache_data)
        rsrc = stack.defn['SwiftContainer']
        self.assertEqual('xyz_convg', rsrc.FnGetRefId())

    @mock.patch('swiftclient.client.Connection.head_account')
    @mock.patch('swiftclient.client.Connection.head_container')
    @mock.patch('swiftclient.client.Connection.put_container')
    def test_parse_live_resource_data(self, mock_put, mock_container,
                                      mock_account):
        stack = utils.parse_stack(self.t)
        container = self._create_container(
            stack, definition_name="SwiftContainerWebsite")
        mock_container.return_value = {
            'x-container-read': '.r:*',
            'x-container-meta-web-index': 'index.html',
            'x-container-meta-web-error': 'error.html',
            'x-container-meta-login': 'login.html'
        }
        mock_account.return_value = {}
        live_state = container.parse_live_resource_data(
            container.properties, container.get_live_resource_data())
        # live state properties values should be equal to current resource
        # properties values
        self.assertEqual(dict(container.properties.items()), live_state)



#    Copyright 2015 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine.resources.openstack.heat import none_resource
from heat.engine import support


class BaseSenlinResource(none_resource.NoneResource):
    """A base class for Senlin resources."""

    support_status = support.SupportStatus(
        version='23.0.0',
        status=support.HIDDEN,
        message=_('Senlin project was retired'),
        previous_status=support.SupportStatus(
            version='22.0.0',
            status=support.DEPRECATED,
            message=_('Senlin project was marked inactive'),
            previous_status=support.SupportStatus(
                version='6.0.0',
                )))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support
from heat.engine import translation


class NeutronQuota(neutron.NeutronResource):
    """A resource for managing neutron quotas.

    Neutron Quota is used to manage operational limits for projects. Currently,
    this resource can manage Neutron's quotas for:

        - subnet
        - network
        - floatingip
        - security_group_rule
        - security_group
        - router
        - port
        - subnetpool
        - rbac_policy

    Note that default neutron security policy usage of this resource
    is limited to being used by administrators only. Administrators should be
    careful to create only one Neutron Quota resource per project, otherwise
    it will be hard for them to manage the quota properly.
    """

    support_status = support.SupportStatus(version='8.0.0')

    required_service_extension = 'quotas'

    PROPERTIES = (
        PROJECT, SUBNET, NETWORK, FLOATINGIP, SECURITY_GROUP_RULE,
        SECURITY_GROUP, ROUTER, PORT, SUBNETPOOL, RBAC_POLICY
    ) = (
        'project', 'subnet', 'network', 'floatingip', 'security_group_rule',
        'security_group', 'router', 'port', 'subnetpool', 'rbac_policy'
    )

    properties_schema = {
        PROJECT: properties.Schema(
            properties.Schema.STRING,
            _('Name or id of the project to set the quota for.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('keystone.project')
            ]
        ),
        SUBNET: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of subnets. '
              'Setting -1 means unlimited.'),
            constraints=[constraints.Range(min=-1)],
            update_allowed=True
        ),
        NETWORK: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of networks. '
              'Setting -1 means unlimited.'),
            constraints=[constraints.Range(min=-1)],
            update_allowed=True
        ),
        FLOATINGIP: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of floating IPs. '
              'Setting -1 means unlimited.'),
            constraints=[constraints.Range(min=-1)],
            update_allowed=True
        ),
        SECURITY_GROUP_RULE: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of security group rules. '
              'Setting -1 means unlimited.'),
            constraints=[constraints.Range(min=-1)],
            update_allowed=True
        ),
        SECURITY_GROUP: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of security groups. '
              'Setting -1 means unlimited.'),
            constraints=[constraints.Range(min=-1)],
            update_allowed=True
        ),
        ROUTER: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of routers. '
              'Setting -1 means unlimited.'),
            constraints=[constraints.Range(min=-1)],
            update_allowed=True
        ),
        PORT: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of ports. '
              'Setting -1 means unlimited.'),
            constraints=[constraints.Range(min=-1)],
            update_allowed=True
        ),
        SUBNETPOOL: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of subnet pools. '
              'Setting -1 means unlimited.'),
            constraints=[constraints.Range(min=-1)],
            update_allowed=True,
            support_status=support.SupportStatus(version='12.0.0')
        ),
        RBAC_POLICY: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of rbac policies. '
              'Setting -1 means unlimited.'),
            constraints=[constraints.Range(min=-1)],
            update_allowed=True,
            support_status=support.SupportStatus(version='12.0.0')
        )
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.PROJECT],
                client_plugin=self.client_plugin('keystone'),
                finder='get_project_id')
        ]

    def handle_create(self):
        self._set_quota()
        self.resource_id_set(self.physical_resource_name())

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self._set_quota(json_snippet.properties(self.properties_schema,
                                                self.context))

    def _set_quota(self, props=None):
        if props is None:
            props = self.properties

        kwargs = dict((k, v) for k, v in props.items()
                      if k != self.PROJECT and v is not None)
        body = {"quota": kwargs}
        self.client().update_quota(props[self.PROJECT], body)

    def handle_delete(self):
        if self.resource_id is not None:
            with self.client_plugin().ignore_not_found:
                self.client().delete_quota(self.resource_id)

    def validate(self):
        super(NeutronQuota, self).validate()
        if sum(1 for p in self.properties.values() if p is not None) <= 1:
            raise exception.PropertyUnspecifiedError(
                *sorted(set(self.PROPERTIES) - {self.PROJECT}))


def resource_mapping():
    return {
        'OS::Neutron::Quota': NeutronQuota
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import itertools
import sys

from oslo_config import cfg
from oslo_log import log

from heat.common import plugin_loader

LOG = log.getLogger(__name__)


class PluginManager(object):
    """A class for managing plugin modules."""

    def __init__(self, *extra_packages):
        """Initialise the Heat Engine plugin package, and any others.

        The heat.engine.plugins package is always created, if it does not
        exist, from the plugin directories specified in the config file, and
        searched for modules. In addition, any extra packages specified are
        also searched for modules. e.g.

        >>> PluginManager('heat.engine.resources')

        will load all modules in the heat.engine.resources package as well as
        any user-supplied plugin modules.
        """
        def packages():
            for package_name in extra_packages:
                yield sys.modules[package_name]

            cfg.CONF.import_opt('plugin_dirs', 'heat.common.config')
            yield plugin_loader.create_subpackage(cfg.CONF.plugin_dirs,
                                                  'heat.engine')

        def modules():
            pkg_modules = map(plugin_loader.load_modules, packages())
            return itertools.chain.from_iterable(pkg_modules)

        self.modules = list(modules())

    def map_to_modules(self, function):
        """Iterate over the results of calling a function on every module."""
        return map(function, self.modules)


class PluginMapping(object):
    """A class for managing plugin mappings."""

    def __init__(self, names, *args, **kwargs):
        """Initialise with the mapping name(s) and arguments.

        `names` can be a single name or a list of names. The first name found
        in a given module is the one used. Each module is searched for a
        function called <name>_mapping() which is called to retrieve the
        mappings provided by that module. Any other arguments passed will be
        passed to the mapping functions.
        """
        if isinstance(names, str):
            names = [names]

        self.names = ['%s_mapping' % name for name in names]
        self.args = args
        self.kwargs = kwargs

    def load_from_module(self, module):
        """Return the mapping specified in the given module.

        If no such mapping is specified, an empty dictionary is returned.
        """
        for mapping_name in self.names:
            mapping_func = getattr(module, mapping_name, None)
            if callable(mapping_func):
                fmt_data = {'mapping_name': mapping_name, 'module': module}
                try:
                    mapping_dict = mapping_func(*self.args, **self.kwargs)
                except Exception:
                    LOG.error('Failed to load %(mapping_name)s '
                              'from %(module)s', fmt_data)
                    raise
                else:
                    if isinstance(mapping_dict, collections.abc.Mapping):
                        return mapping_dict
                    elif mapping_dict is not None:
                        LOG.error('Invalid type for %(mapping_name)s '
                                  'from %(module)s', fmt_data)

        return {}

    def load_all(self, plugin_manager):
        """Iterate over the mappings from all modules in the plugin manager.

        Mappings are returned as a list of (key, value) tuples.
        """
        mod_dicts = plugin_manager.map_to_modules(self.load_from_module)
        return itertools.chain.from_iterable(d.items() for d
                                             in mod_dicts)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import uuid

from oslo_config import cfg
from oslo_messaging.rpc import dispatcher
from oslo_serialization import jsonutils as json

from heat.common import context
from heat.common import environment_util as env_util
from heat.common import exception
from heat.common import identifier
from heat.common import policy
from heat.common import template_format
from heat.engine.cfn import template as cfntemplate
from heat.engine import environment
from heat.engine.hot import functions as hot_functions
from heat.engine.hot import template as hottemplate
from heat.engine import resource as res
from heat.engine import service
from heat.engine import stack as parser
from heat.engine import template as templatem
from heat.objects import stack as stack_object
from heat.rpc import api as rpc_api
from heat.tests import common
from heat.tests.engine import tools
from heat.tests import generic_resource as generic_rsrc
from heat.tests.openstack.nova import fakes as fakes_nova
from heat.tests import utils

cfg.CONF.import_opt('engine_life_check_timeout', 'heat.common.config')
cfg.CONF.import_opt('enable_stack_abandon', 'heat.common.config')

wp_template_no_default = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "WordPress",
  "Parameters" : {
    "KeyName" : {
      "Description" : "KeyName",
      "Type" : "String"
    }
  },
  "Resources" : {
    "WebServer": {
      "Type": "AWS::EC2::Instance",
      "Properties": {
        "ImageId" : "F17-x86_64-gold",
        "InstanceType"   : "m1.large",
        "KeyName"        : "test",
        "UserData"       : "wordpress"
      }
    }
  }
}
'''

user_policy_template = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "Just a User",
  "Parameters" : {},
  "Resources" : {
    "CfnUser" : {
      "Type" : "AWS::IAM::User",
      "Properties" : {
        "Policies" : [ { "Ref": "WebServerAccessPolicy"} ]
      }
    },
    "WebServerAccessPolicy" : {
      "Type" : "OS::Heat::AccessPolicy",
      "Properties" : {
        "AllowedResources" : [ "WebServer" ]
      }
    },
    "HostKeys" : {
      "Type" : "AWS::IAM::AccessKey",
      "Properties" : {
        "UserName" : {"Ref": "CfnUser"}
      }
    },
    "WebServer": {
      "Type": "AWS::EC2::Instance",
      "Properties": {
        "ImageId" : "F17-x86_64-gold",
        "InstanceType"   : "m1.large",
        "KeyName"        : "test",
        "UserData"       : "wordpress"
      }
    }
  }
}
'''

server_config_template = '''
heat_template_version: 2013-05-23
resources:
  WebServer:
    type: OS::Nova::Server
'''


class StackCreateTest(common.HeatTestCase):
    def test_wordpress_single_instance_stack_create(self):
        stack = tools.get_stack('test_stack', utils.dummy_context())
        fc = tools.setup_mocks_with_mock(self, stack)
        stack.store()
        stack.create()

        self.assertIsNotNone(stack['WebServer'])
        self.assertGreater(int(stack['WebServer'].resource_id), 0)
        self.assertNotEqual(stack['WebServer'].ipaddress, '0.0.0.0')
        tools.validate_setup_mocks_with_mock(stack, fc)

    def test_wordpress_single_instance_stack_adopt(self):
        t = template_format.parse(tools.wp_template)
        template = templatem.Template(t)
        ctx = utils.dummy_context()
        adopt_data = {
            'resources': {
                'WebServer': {
                    'resource_id': 'test-res-id'
                }
            }
        }
        stack = parser.Stack(ctx,
                             'test_stack',
                             template,
                             adopt_stack_data=adopt_data)

        fc = tools.setup_mocks_with_mock(self, stack,
                                         mock_image_constraint=False)
        stack.store()
        stack.adopt()

        self.assertIsNotNone(stack['WebServer'])
        self.assertEqual('test-res-id', stack['WebServer'].resource_id)
        self.assertEqual((stack.ADOPT, stack.COMPLETE), stack.state)
        tools.validate_setup_mocks_with_mock(
            stack, fc, mock_image_constraint=False, validate_create=False)

    def test_wordpress_single_instance_stack_adopt_fail(self):
        t = template_format.parse(tools.wp_template)
        template = templatem.Template(t)
        ctx = utils.dummy_context()
        adopt_data = {
            'resources': {
                'WebServer1': {
                    'resource_id': 'test-res-id'
                }
            }
        }
        stack = parser.Stack(ctx,
                             'test_stack',
                             template,
                             adopt_stack_data=adopt_data)

        fc = tools.setup_mocks_with_mock(self, stack,
                                         mock_image_constraint=False)
        stack.store()
        stack.adopt()
        self.assertIsNotNone(stack['WebServer'])
        expected = ('Resource ADOPT failed: Exception: resources.WebServer: '
                    'Resource ID was not provided.')
        self.assertEqual(expected, stack.status_reason)
        self.assertEqual((stack.ADOPT, stack.FAILED), stack.state)
        tools.validate_setup_mocks_with_mock(
            stack, fc, mock_image_constraint=False, validate_create=False)

    def test_wordpress_single_instance_stack_delete(self):
        ctx = utils.dummy_context()
        stack = tools.get_stack('test_stack', ctx)
        fc = tools.setup_mocks_with_mock(self, stack, mock_keystone=False)
        stack_id = stack.store()
        stack.create()

        db_s = stack_object.Stack.get_by_id(ctx, stack_id)
        self.assertIsNotNone(db_s)

        self.assertIsNotNone(stack['WebServer'])
        self.assertGreater(int(stack['WebServer'].resource_id), 0)

        self.patchobject(fc.servers, 'delete',
                         side_effect=fakes_nova.fake_exception())
        stack.delete()

        rsrc = stack['WebServer']
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)
        self.assertEqual((stack.DELETE, stack.COMPLETE), rsrc.state)
        self.assertIsNone(stack_object.Stack.get_by_id(ctx, stack_id))

        db_s.refresh()
        self.assertEqual('DELETE', db_s.action)
        self.assertEqual('COMPLETE', db_s.status)
        tools.validate_setup_mocks_with_mock(stack, fc)


class StackConvergenceServiceCreateUpdateTest(common.HeatTestCase):

    def setUp(self):
        super(StackConvergenceServiceCreateUpdateTest, self).setUp()
        cfg.CONF.set_override('convergence_engine', True)
        self.ctx = utils.dummy_context()
        self.man = service.EngineService('a-host', 'a-topic')
        self.man.thread_group_mgr = tools.DummyThreadGroupManager()

    def _stub_update_mocks(self, stack_to_return):
        self.patchobject(parser, 'Stack')
        parser.Stack.load.return_value = stack_to_return

        self.patchobject(templatem, 'Template')
        self.patchobject(environment, 'Environment')

    def _test_stack_create_convergence(self, stack_name):
        params = {'foo': 'bar'}
        template = '{ "Template": "data" }'

        stack = tools.get_stack(stack_name, self.ctx,
                                template=tools.string_template_five,
                                convergence=True)

        stack.converge = None
        self.patchobject(templatem, 'Template', return_value=stack.t)
        self.patchobject(environment, 'Environment', return_value=stack.env)
        self.patchobject(parser, 'Stack', return_value=stack)
        self.patchobject(stack, 'validate', return_value=None)

        api_args = {'timeout_mins': 60, 'disable_rollback': False}
        result = self.man.create_stack(self.ctx, 'service_create_test_stack',
                                       template, params, None, api_args)
        db_stack = stack_object.Stack.get_by_id(self.ctx, result['stack_id'])
        self.assertTrue(db_stack.convergence)
        self.assertEqual(result['stack_id'], db_stack.id)
        templatem.Template.assert_called_once_with(template, files=None)
        environment.Environment.assert_called_once_with(params)
        parser.Stack.assert_called_once_with(
            self.ctx, stack.name,
            stack.t, owner_id=None,
            parent_resource=None,
            nested_depth=0, user_creds_id=None,
            stack_user_project_id=None,
            timeout_mins=60,
            disable_rollback=False,
            convergence=True)

    def test_stack_create_enabled_convergence_engine(self):
        stack_name = 'service_create_test_stack'
        self._test_stack_create_convergence(stack_name)

    def test_stack_update_enabled_convergence_engine(self):
        stack_name = 'service_update_test_stack'
        params = {'foo': 'bar'}
        template = '{ "Template": "data" }'
        old_stack = tools.get_stack(stack_name, self.ctx,
                                    template=tools.string_template_five,
                                    convergence=True)
        old_stack.timeout_mins = 1
        old_stack.store()
        stack = tools.get_stack(stack_name, self.ctx,
                                template=tools.string_template_five_update,
                                convergence=True)

        self._stub_update_mocks(old_stack)

        templatem.Template.return_value = stack.t
        environment.Environment.return_value = stack.env
        parser.Stack.return_value = stack

        self.patchobject(stack, 'validate', return_value=None)

        api_args = {'timeout_mins': 60, 'disable_rollback': False,
                    rpc_api.PARAM_CONVERGE: False}
        result = self.man.update_stack(self.ctx, old_stack.identifier(),
                                       template, params, None, api_args)
        self.assertTrue(old_stack.convergence)
        self.assertEqual(old_stack.identifier(), result)
        self.assertIsInstance(result, dict)
        self.assertTrue(result['stack_id'])
        parser.Stack.load.assert_called_once_with(
            self.ctx, stack=mock.ANY, check_refresh_cred=True)
        templatem.Template.assert_called_once_with(template, files=None)
        environment.Environment.assert_called_once_with(params)


class StackServiceAuthorizeTest(common.HeatTestCase):

    def setUp(self):
        super(StackServiceAuthorizeTest, self).setUp()

        self.ctx = utils.dummy_context(tenant_id='stack_service_test_tenant')
        self.eng = service.EngineService('a-host', 'a-topic')
        self.eng.engine_id = 'engine-fake-uuid'

    @tools.stack_context('service_authorize_stack_user_nocreds_test_stack')
    def test_stack_authorize_stack_user_nocreds(self):
        self.assertFalse(self.eng._authorize_stack_user(self.ctx,
                                                        self.stack,
                                                        'foo'))

    @tools.stack_context('service_authorize_user_attribute_error_test_stack')
    def test_stack_authorize_stack_user_attribute_error(self):
        self.patchobject(json, 'loads', side_effect=AttributeError)
        self.assertFalse(self.eng._authorize_stack_user(self.ctx,
                                                        self.stack,
                                                        'foo'))
        json.loads.assert_called_once_with(None)

    @tools.stack_context('service_authorize_stack_user_type_error_test_stack')
    def test_stack_authorize_stack_user_type_error(self):
        self.patchobject(json, 'loads', side_effect=TypeError)
        self.assertFalse(self.eng._authorize_stack_user(self.ctx,
                                                        self.stack,
                                                        'foo'))
        json.loads.assert_called_once_with(None)

    def test_stack_authorize_stack_user(self):
        self.ctx = utils.dummy_context()
        self.ctx.aws_creds = '{"ec2Credentials": {"access": "4567"}}'
        stack_name = 'stack_authorize_stack_user'
        stack = tools.get_stack(stack_name, self.ctx, user_policy_template)
        self.stack = stack
        fc = tools.setup_mocks_with_mock(self, stack)
        self.patchobject(fc.servers, 'delete',
                         side_effect=fakes_nova.fake_exception())

        stack.store()
        stack.create()

        self.assertTrue(self.eng._authorize_stack_user(
            self.ctx, self.stack, 'WebServer'))

        self.assertFalse(self.eng._authorize_stack_user(
            self.ctx, self.stack, 'CfnUser'))

        self.assertFalse(self.eng._authorize_stack_user(
            self.ctx, self.stack, 'NoSuchResource'))
        tools.validate_setup_mocks_with_mock(stack, fc)

    def test_stack_authorize_stack_user_user_id(self):
        self.ctx = utils.dummy_context(user_id=str(uuid.uuid4()))
        stack_name = 'stack_authorize_stack_user_user_id'
        stack = tools.get_stack(stack_name, self.ctx, server_config_template)
        self.stack = stack

        def handler(resource_name):
            return resource_name == 'WebServer'

        self.stack.register_access_allowed_handler(self.ctx.user_id, handler)

        # matching credential_id and resource_name
        self.assertTrue(self.eng._authorize_stack_user(
            self.ctx, self.stack, 'WebServer'))

        # not matching resource_name
        self.assertFalse(self.eng._authorize_stack_user(
            self.ctx, self.stack, 'NoSuchResource'))

        # not matching credential_id
        self.ctx.user = str(uuid.uuid4())
        self.assertFalse(self.eng._authorize_stack_user(
            self.ctx, self.stack, 'WebServer'))


class StackServiceTest(common.HeatTestCase):

    def setUp(self):
        super(StackServiceTest, self).setUp()

        self.ctx = utils.dummy_context(tenant_id='stack_service_test_tenant')
        self.eng = service.EngineService('a-host', 'a-topic')
        self.eng.thread_group_mgr = tools.DummyThreadGroupManager()
        self.eng.engine_id = 'engine-fake-uuid'

    @tools.stack_context('service_identify_test_stack', False)
    def test_stack_identify(self):
        identity = self.eng.identify_stack(self.ctx, self.stack.name)
        self.assertEqual(self.stack.identifier(), identity)

    @tools.stack_context('ef0c41a4-644f-447c-ad80-7eecb0becf79', False)
    def test_stack_identify_by_name_in_uuid(self):
        identity = self.eng.identify_stack(self.ctx, self.stack.name)
        self.assertEqual(self.stack.identifier(), identity)

    @tools.stack_context('service_identify_uuid_test_stack', False)
    def test_stack_identify_uuid(self):
        identity = self.eng.identify_stack(self.ctx, self.stack.id)
        self.assertEqual(self.stack.identifier(), identity)

    def test_stack_identify_nonexist(self):
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.eng.identify_stack, self.ctx, 'wibble')
        self.assertEqual(exception.EntityNotFound, ex.exc_info[0])

    @tools.stack_context('service_create_existing_test_stack', False)
    def test_stack_create_existing(self):
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.eng.create_stack, self.ctx,
                               self.stack.name, self.stack.t.t, {}, None, {})
        self.assertEqual(exception.StackExists, ex.exc_info[0])

    @tools.stack_context('service_name_tenants_test_stack', False)
    def test_stack_by_name_tenants(self):
        self.assertEqual(
            self.stack.id,
            stack_object.Stack.get_by_name(self.ctx, self.stack.name).id
        )
        ctx2 = utils.dummy_context(tenant_id='stack_service_test_tenant2')
        self.assertIsNone(stack_object.Stack.get_by_name(
            ctx2,
            self.stack.name))

    @tools.stack_context('service_badname_test_stack', False)
    def test_stack_by_name_badname(self):
        # If a bad name type, such as a map, is passed, we should just return
        # None, as it's converted to a string, which won't match any name
        ctx = utils.dummy_context(tenant_id='stack_service_test_tenant')
        self.assertIsNone(stack_object.Stack.get_by_name(
            ctx,
            {'notallowed': self.stack.name}))
        self.assertIsNone(stack_object.Stack.get_by_name_and_owner_id(
            ctx,
            {'notallowed': self.stack.name}, 'owner'))

    @tools.stack_context('service_list_all_test_stack')
    def test_stack_list_all(self):
        sl = self.eng.list_stacks(self.ctx)

        self.assertEqual(1, len(sl))
        for s in sl:
            self.assertIn('creation_time', s)
            self.assertIn('updated_time', s)
            self.assertIn('deletion_time', s)
            self.assertIsNone(s['deletion_time'])
            self.assertIn('stack_identity', s)
            self.assertIsNotNone(s['stack_identity'])
            self.assertIn('stack_name', s)
            self.assertEqual(self.stack.name, s['stack_name'])
            self.assertIn('stack_status', s)
            self.assertIn('stack_status_reason', s)
            self.assertIn('description', s)
            self.assertEqual('', s['description'])

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_passes_marker_info(self, mock_stack_get_all):
        limit = object()
        marker = object()
        sort_keys = object()
        sort_dir = object()
        self.eng.list_stacks(self.ctx, limit=limit, marker=marker,
                             sort_keys=sort_keys, sort_dir=sort_dir)
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=limit,
                                                   sort_keys=sort_keys,
                                                   marker=marker,
                                                   sort_dir=sort_dir,
                                                   filters=mock.ANY,
                                                   show_deleted=mock.ANY,
                                                   show_nested=mock.ANY,
                                                   show_hidden=mock.ANY,
                                                   tags=mock.ANY,
                                                   tags_any=mock.ANY,
                                                   not_tags=mock.ANY,
                                                   not_tags_any=mock.ANY)

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_passes_filtering_info(self, mock_stack_get_all):
        filters = {'foo': 'bar'}
        self.eng.list_stacks(self.ctx, filters=filters)
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=mock.ANY,
                                                   sort_keys=mock.ANY,
                                                   marker=mock.ANY,
                                                   sort_dir=mock.ANY,
                                                   filters=filters,
                                                   show_deleted=mock.ANY,
                                                   show_nested=mock.ANY,
                                                   show_hidden=mock.ANY,
                                                   tags=mock.ANY,
                                                   tags_any=mock.ANY,
                                                   not_tags=mock.ANY,
                                                   not_tags_any=mock.ANY)

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_passes_filter_translated(self, mock_stack_get_all):
        filters = {'stack_name': 'bar'}
        self.eng.list_stacks(self.ctx, filters=filters)
        translated = {'name': 'bar'}
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=mock.ANY,
                                                   sort_keys=mock.ANY,
                                                   marker=mock.ANY,
                                                   sort_dir=mock.ANY,
                                                   filters=translated,
                                                   show_deleted=mock.ANY,
                                                   show_nested=mock.ANY,
                                                   show_hidden=mock.ANY,
                                                   tags=mock.ANY,
                                                   tags_any=mock.ANY,
                                                   not_tags=mock.ANY,
                                                   not_tags_any=mock.ANY)

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_show_nested(self, mock_stack_get_all):
        self.eng.list_stacks(self.ctx, show_nested=True)
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=mock.ANY,
                                                   sort_keys=mock.ANY,
                                                   marker=mock.ANY,
                                                   sort_dir=mock.ANY,
                                                   filters=mock.ANY,
                                                   show_deleted=mock.ANY,
                                                   show_nested=True,
                                                   show_hidden=mock.ANY,
                                                   tags=mock.ANY,
                                                   tags_any=mock.ANY,
                                                   not_tags=mock.ANY,
                                                   not_tags_any=mock.ANY)

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_show_deleted(self, mock_stack_get_all):
        self.eng.list_stacks(self.ctx, show_deleted=True)
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=mock.ANY,
                                                   sort_keys=mock.ANY,
                                                   marker=mock.ANY,
                                                   sort_dir=mock.ANY,
                                                   filters=mock.ANY,
                                                   show_deleted=True,
                                                   show_nested=mock.ANY,
                                                   show_hidden=mock.ANY,
                                                   tags=mock.ANY,
                                                   tags_any=mock.ANY,
                                                   not_tags=mock.ANY,
                                                   not_tags_any=mock.ANY)

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_show_hidden(self, mock_stack_get_all):
        self.eng.list_stacks(self.ctx, show_hidden=True)
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=mock.ANY,
                                                   sort_keys=mock.ANY,
                                                   marker=mock.ANY,
                                                   sort_dir=mock.ANY,
                                                   filters=mock.ANY,
                                                   show_deleted=mock.ANY,
                                                   show_nested=mock.ANY,
                                                   show_hidden=True,
                                                   tags=mock.ANY,
                                                   tags_any=mock.ANY,
                                                   not_tags=mock.ANY,
                                                   not_tags_any=mock.ANY)

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_tags(self, mock_stack_get_all):
        self.eng.list_stacks(self.ctx, tags=['foo', 'bar'])
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=mock.ANY,
                                                   sort_keys=mock.ANY,
                                                   marker=mock.ANY,
                                                   sort_dir=mock.ANY,
                                                   filters=mock.ANY,
                                                   show_deleted=mock.ANY,
                                                   show_nested=mock.ANY,
                                                   show_hidden=mock.ANY,
                                                   tags=['foo', 'bar'],
                                                   tags_any=mock.ANY,
                                                   not_tags=mock.ANY,
                                                   not_tags_any=mock.ANY)

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_tags_any(self, mock_stack_get_all):
        self.eng.list_stacks(self.ctx, tags_any=['foo', 'bar'])
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=mock.ANY,
                                                   sort_keys=mock.ANY,
                                                   marker=mock.ANY,
                                                   sort_dir=mock.ANY,
                                                   filters=mock.ANY,
                                                   show_deleted=mock.ANY,
                                                   show_nested=mock.ANY,
                                                   show_hidden=mock.ANY,
                                                   tags=mock.ANY,
                                                   tags_any=['foo', 'bar'],
                                                   not_tags=mock.ANY,
                                                   not_tags_any=mock.ANY)

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_not_tags(self, mock_stack_get_all):
        self.eng.list_stacks(self.ctx, not_tags=['foo', 'bar'])
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=mock.ANY,
                                                   sort_keys=mock.ANY,
                                                   marker=mock.ANY,
                                                   sort_dir=mock.ANY,
                                                   filters=mock.ANY,
                                                   show_deleted=mock.ANY,
                                                   show_nested=mock.ANY,
                                                   show_hidden=mock.ANY,
                                                   tags=mock.ANY,
                                                   tags_any=mock.ANY,
                                                   not_tags=['foo', 'bar'],
                                                   not_tags_any=mock.ANY)

    @mock.patch.object(stack_object.Stack, 'get_all')
    def test_stack_list_not_tags_any(self, mock_stack_get_all):
        self.eng.list_stacks(self.ctx, not_tags_any=['foo', 'bar'])
        mock_stack_get_all.assert_called_once_with(self.ctx,
                                                   limit=mock.ANY,
                                                   sort_keys=mock.ANY,
                                                   marker=mock.ANY,
                                                   sort_dir=mock.ANY,
                                                   filters=mock.ANY,
                                                   show_deleted=mock.ANY,
                                                   show_nested=mock.ANY,
                                                   show_hidden=mock.ANY,
                                                   tags=mock.ANY,
                                                   tags_any=mock.ANY,
                                                   not_tags=mock.ANY,
                                                   not_tags_any=['foo', 'bar'])

    @mock.patch.object(stack_object.Stack, 'count_all')
    def test_count_stacks_passes_filter_info(self, mock_stack_count_all):
        self.eng.count_stacks(self.ctx, filters={'foo': 'bar'})
        mock_stack_count_all.assert_called_once_with(mock.ANY,
                                                     filters={'foo': 'bar'},
                                                     show_deleted=False,
                                                     show_nested=False,
                                                     show_hidden=False,
                                                     tags=None,
                                                     tags_any=None,
                                                     not_tags=None,
                                                     not_tags_any=None)

    @mock.patch.object(stack_object.Stack, 'count_all')
    def test_count_stacks_show_nested(self, mock_stack_count_all):
        self.eng.count_stacks(self.ctx, show_nested=True)
        mock_stack_count_all.assert_called_once_with(mock.ANY,
                                                     filters=mock.ANY,
                                                     show_deleted=False,
                                                     show_nested=True,
                                                     show_hidden=False,
                                                     tags=None,
                                                     tags_any=None,
                                                     not_tags=None,
                                                     not_tags_any=None)

    @mock.patch.object(stack_object.Stack, 'count_all')
    def test_count_stack_show_deleted(self, mock_stack_count_all):
        self.eng.count_stacks(self.ctx, show_deleted=True)
        mock_stack_count_all.assert_called_once_with(mock.ANY,
                                                     filters=mock.ANY,
                                                     show_deleted=True,
                                                     show_nested=False,
                                                     show_hidden=False,
                                                     tags=None,
                                                     tags_any=None,
                                                     not_tags=None,
                                                     not_tags_any=None)

    @mock.patch.object(stack_object.Stack, 'count_all')
    def test_count_stack_show_hidden(self, mock_stack_count_all):
        self.eng.count_stacks(self.ctx, show_hidden=True)
        mock_stack_count_all.assert_called_once_with(mock.ANY,
                                                     filters=mock.ANY,
                                                     show_deleted=False,
                                                     show_nested=False,
                                                     show_hidden=True,
                                                     tags=None,
                                                     tags_any=None,
                                                     not_tags=None,
                                                     not_tags_any=None)

    @tools.stack_context('service_export_stack')
    def test_export_stack(self):
        cfg.CONF.set_override('enable_stack_abandon', True)
        self.patchobject(parser.Stack, 'load', return_value=self.stack)
        expected_res = {
            'WebServer': {
                'action': 'CREATE',
                'metadata': {},
                'name': 'WebServer',
                'resource_data': {},
                'resource_id': '9999',
                'status': 'COMPLETE',
                'type': 'AWS::EC2::Instance'}}
        self.stack.tags = ['tag1', 'tag2']
        ret = self.eng.export_stack(self.ctx, self.stack.identifier())
        self.assertEqual(11, len(ret))
        self.assertEqual('CREATE', ret['action'])
        self.assertEqual('COMPLETE', ret['status'])
        self.assertEqual('service_export_stack', ret['name'])
        self.assertEqual({}, ret['files'])
        self.assertIn('id', ret)
        self.assertEqual(expected_res, ret['resources'])
        self.assertEqual(self.stack.t.t, ret['template'])
        self.assertIn('project_id', ret)
        self.assertIn('stack_user_project_id', ret)
        self.assertIn('environment', ret)
        self.assertIn('files', ret)
        self.assertEqual(['tag1', 'tag2'], ret['tags'])

    @tools.stack_context('service_abandon_stack')
    def test_abandon_stack(self):
        cfg.CONF.set_override('enable_stack_abandon', True)
        self.patchobject(parser.Stack, 'load', return_value=self.stack)
        self.eng.abandon_stack(self.ctx, self.stack.identifier())
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.eng.show_stack,
                               self.ctx, self.stack.identifier(),
                               resolve_outputs=True)
        self.assertEqual(exception.EntityNotFound, ex.exc_info[0])

    def test_stack_describe_nonexistent(self):
        non_exist_identifier = identifier.HeatIdentifier(
            self.ctx.project_id, 'wibble',
            '18d06e2e-44d3-4bef-9fbf-52480d604b02')

        stack_not_found_exc = exception.EntityNotFound(
            entity='Stack', name='test')
        self.patchobject(service.EngineService, '_get_stack',
                         side_effect=stack_not_found_exc)

        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.eng.show_stack,
                               self.ctx, non_exist_identifier,
                               resolve_outputs=True)
        self.assertEqual(exception.EntityNotFound, ex.exc_info[0])
        service.EngineService._get_stack.assert_called_once_with(
            self.ctx, non_exist_identifier,
            show_deleted=True)

    def test_stack_describe_bad_tenant(self):
        non_exist_identifier = identifier.HeatIdentifier(
            'wibble', 'wibble',
            '18d06e2e-44d3-4bef-9fbf-52480d604b02')

        invalid_tenant_exc = exception.InvalidTenant(target='test',
                                                     actual='test')
        self.patchobject(service.EngineService, '_get_stack',
                         side_effect=invalid_tenant_exc)

        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.eng.show_stack,
                               self.ctx, non_exist_identifier,
                               resolve_outputs=True)
        self.assertEqual(exception.InvalidTenant, ex.exc_info[0])
        service.EngineService._get_stack.assert_called_once_with(
            self.ctx, non_exist_identifier,
            show_deleted=True)

    @tools.stack_context('service_describe_test_stack', False)
    def test_stack_describe(self):
        s = stack_object.Stack.get_by_id(self.ctx, self.stack.id)
        self.patchobject(service.EngineService, '_get_stack', return_value=s)

        sl = self.eng.show_stack(self.ctx, self.stack.identifier(),
                                 resolve_outputs=True)

        self.assertEqual(1, len(sl))

        s = sl[0]
        self.assertIn('creation_time', s)
        self.assertIn('updated_time', s)
        self.assertIn('deletion_time', s)
        self.assertIsNone(s['deletion_time'])
        self.assertIn('stack_identity', s)
        self.assertIsNotNone(s['stack_identity'])
        self.assertIn('stack_name', s)
        self.assertEqual(self.stack.name, s['stack_name'])
        self.assertIn('stack_status', s)
        self.assertIn('stack_status_reason', s)
        self.assertIn('description', s)
        self.assertIn('WordPress', s['description'])
        self.assertIn('parameters', s)
        service.EngineService._get_stack.assert_called_once_with(
            self.ctx,
            self.stack.identifier(),
            show_deleted=True)

    @tools.stack_context('service_describe_all_test_stack', False)
    def test_stack_describe_all(self):
        sl = self.eng.show_stack(self.ctx, None, resolve_outputs=True)

        self.assertEqual(1, len(sl))

        s = sl[0]
        self.assertIn('creation_time', s)
        self.assertIn('updated_time', s)
        self.assertIn('deletion_time', s)
        self.assertIsNone(s['deletion_time'])
        self.assertIn('stack_identity', s)
        self.assertIsNotNone(s['stack_identity'])
        self.assertIn('stack_name', s)
        self.assertEqual(self.stack.name, s['stack_name'])
        self.assertIn('stack_status', s)
        self.assertIn('stack_status_reason', s)
        self.assertIn('description', s)
        self.assertIn('WordPress', s['description'])
        self.assertIn('parameters', s)

    @mock.patch('heat.engine.template._get_template_extension_manager')
    def test_list_template_versions(self, templ_mock):

        class DummyMgr(object):
            def names(self):
                return ['a.2012-12-12', 'c.newton', 'c.2016-10-14',
                        'c.something']

            def __getitem__(self, item):
                m = mock.MagicMock()
                if item == 'a.2012-12-12':
                    m.plugin = cfntemplate.CfnTemplate
                    return m
                else:
                    m.plugin = hottemplate.HOTemplate20130523
                    return m

        templ_mock.return_value = DummyMgr()
        templates = self.eng.list_template_versions(self.ctx)
        expected = [{'version': 'a.2012-12-12', 'type': 'cfn', 'aliases': []},
                    {'version': 'c.2016-10-14',
                     'aliases': ['c.newton', 'c.something'], 'type': 'hot'}]
        self.assertEqual(expected, templates)

    @mock.patch('heat.engine.template._get_template_extension_manager')
    def test_list_template_versions_invalid_version(self, templ_mock):

        class DummyMgr(object):
            def names(self):
                return ['c.something']

            def __getitem__(self, item):
                m = mock.MagicMock()
                if item == 'c.something':
                    m.plugin = cfntemplate.CfnTemplate
                    return m

        templ_mock.return_value = DummyMgr()
        ret = self.assertRaises(exception.InvalidTemplateVersions,
                                self.eng.list_template_versions, self.ctx)
        self.assertIn('A template version alias c.something was added',
                      str(ret))

    @mock.patch('heat.engine.template._get_template_extension_manager')
    def test_list_template_functions(self, templ_mock):

        class DummyFunc1(object):
            """Dummy Func1.

            Dummy Func1 Long Description.
            """

        class DummyFunc2(object):
            """Dummy Func2.

            Dummy Func2 Long Description.
            """

        class DummyConditionFunc(object):
            """Dummy Condition Func.

            Dummy Condition Func Long Description.
            """

        plugin_mock = mock.Mock(
            functions={'dummy1': DummyFunc1,
                       'dummy2': DummyFunc2,
                       'removed': hot_functions.Removed},
            condition_functions={'condition_dummy': DummyConditionFunc})
        dummy_tmpl = mock.Mock(plugin=plugin_mock)

        class DummyMgr(object):
            def __getitem__(self, item):
                return dummy_tmpl

        templ_mock.return_value = DummyMgr()
        functions = self.eng.list_template_functions(self.ctx, 'dummytemplate')
        expected = [{'functions': 'dummy1',
                     'description': 'Dummy Func1.'},
                    {'functions': 'dummy2',
                     'description': 'Dummy Func2.'}]
        self.assertEqual(sorted(expected, key=lambda k: k['functions']),
                         sorted(functions, key=lambda k: k['functions']))

        # test with_condition
        functions = self.eng.list_template_functions(self.ctx, 'dummytemplate',
                                                     with_condition=True)
        expected = [{'functions': 'dummy1',
                     'description': 'Dummy Func1.'},
                    {'functions': 'dummy2',
                     'description': 'Dummy Func2.'},
                    {'functions': 'condition_dummy',
                     'description': 'Dummy Condition Func.'}]
        self.assertEqual(sorted(expected, key=lambda k: k['functions']),
                         sorted(functions, key=lambda k: k['functions']))

    @mock.patch('heat.engine.template._get_template_extension_manager')
    def test_list_template_functions_version_not_found(self, templ_mock):
        class DummyMgr(object):
            def __getitem__(self, item):
                raise KeyError()

        templ_mock.return_value = DummyMgr()
        version = 'dummytemplate'
        ex = self.assertRaises(exception.NotFound,
                               self.eng.list_template_functions,
                               self.ctx,
                               version)
        msg = "Template with version %s not found" % version
        self.assertEqual(msg, str(ex))

    def test_stack_list_outputs(self):
        t = template_format.parse(tools.wp_template)
        t['outputs'] = {
            'test': {'value': '{ get_attr: fir }',
                     'description': 'sec'},
            'test2': {'value': 'sec'}}
        tmpl = templatem.Template(t)
        stack = parser.Stack(self.ctx, 'service_list_outputs_stack', tmpl)

        self.patchobject(self.eng, '_get_stack')
        self.patchobject(parser.Stack, 'load', return_value=stack)

        outputs = self.eng.list_outputs(self.ctx, mock.ANY)

        self.assertIn({'output_key': 'test',
                       'description': 'sec'}, outputs)
        self.assertIn({'output_key': 'test2',
                       'description': 'No description given'},
                      outputs)

    def test_stack_empty_list_outputs(self):
        # Ensure that stack with no output returns empty list
        t = template_format.parse(tools.wp_template)
        t['outputs'] = {}
        tmpl = templatem.Template(t)
        stack = parser.Stack(self.ctx, 'service_list_outputs_stack', tmpl)

        self.patchobject(self.eng, '_get_stack')
        self.patchobject(parser.Stack, 'load', return_value=stack)

        outputs = self.eng.list_outputs(self.ctx, mock.ANY)
        self.assertEqual([], outputs)

    def test_stack_delete_complete_is_not_found(self):
        t = template_format.parse(tools.wp_template)
        tmpl = templatem.Template(t)
        stack = parser.Stack(self.ctx, 'delete_complete_stack', tmpl)
        self.patchobject(self.eng, '_get_stack')
        self.patchobject(parser.Stack, 'load', return_value=stack)
        stack.status = stack.COMPLETE
        stack.action = stack.DELETE
        stack.convergence = True
        self.eng.thread_group_mgr.start = mock.MagicMock()
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.eng.delete_stack,
                               'irrelevant',
                               'irrelevant')
        self.assertEqual(exception.EntityNotFound, ex.exc_info[0])
        self.eng.thread_group_mgr.start.assert_called_once_with(
            None, stack.purge_db)

    def test_get_environment(self):
        # Setup
        t = template_format.parse(tools.wp_template)
        env = {'parameters': {'KeyName': 'EnvKey'}}
        tmpl = templatem.Template(t)
        stack = parser.Stack(self.ctx, 'get_env_stack', tmpl)
        stack.store()

        mock_get_stack = self.patchobject(self.eng, '_get_stack')
        mock_get_stack.return_value = mock.MagicMock()
        mock_get_stack.return_value.raw_template.environment = env
        self.patchobject(templatem.Template, 'load', return_value=tmpl)

        # Test
        found = self.eng.get_environment(self.ctx, stack.identifier())

        # Verify
        self.assertEqual(env, found)

    def test_get_environment_hidden_param(self):
        # Setup
        env = {
            'parameters': {
                'admin': 'testuser',
                'pass': 'pa55w0rd'
            },
            'parameter_defaults': {
                'secret': 'dummy'
            },
            'resource_registry': {
                'res': 'resource.yaml'
            }
        }
        t = {
            'heat_template_version': '2018-08-31',
            'parameters': {
                'admin': {'type': 'string'},
                'pass': {'type': 'string', 'hidden': True}
            },
            'resources': {
                'res1': {'type': 'res'}
            }
        }
        files = {
            'resource.yaml': '''
                heat_template_version: 2018-08-31
                parameters:
                    secret:
                        type: string
                        hidden: true
            '''
        }
        tmpl = templatem.Template(t, files=files)
        stack = parser.Stack(self.ctx, 'get_env_stack', tmpl)
        stack.store()

        mock_get_stack = self.patchobject(self.eng, '_get_stack')
        mock_get_stack.return_value = mock.MagicMock()
        mock_get_stack.return_value.raw_template.environment = env
        self.patchobject(templatem.Template, 'load', return_value=tmpl)

        # Test
        found = self.eng.get_environment(self.ctx, stack.identifier())

        # Verify
        env['parameters']['pass'] = '******'
        env['parameter_defaults']['secret'] = '******'
        self.assertEqual(env, found)

    def test_get_environment_no_env(self):
        # Setup
        exc = exception.EntityNotFound(entity='stack', name='missing')
        self.patchobject(self.eng, '_get_stack', side_effect=exc)

        # Test
        self.assertRaises(dispatcher.ExpectedException,
                          self.eng.get_environment,
                          self.ctx,
                          'irrelevant')

    def test_get_files(self):
        # Setup
        t = template_format.parse(tools.wp_template)
        files = {'foo.yaml': 'i am a file'}
        tmpl = templatem.Template(t, files=files)
        stack = parser.Stack(self.ctx, 'get_env_stack', tmpl)
        stack.store()

        mock_get_stack = self.patchobject(self.eng, '_get_stack')
        mock_get_stack.return_value = mock.MagicMock()
        self.patchobject(templatem.Template, 'load', return_value=tmpl)

        # Test
        found = self.eng.get_files(self.ctx, stack.identifier())

        # Verify
        self.assertEqual(files, found)

    def test_stack_show_output(self):
        t = template_format.parse(tools.wp_template)
        t['outputs'] = {'test': {'value': 'first', 'description': 'sec'},
                        'test2': {'value': 'sec'}}
        tmpl = templatem.Template(t)
        stack = parser.Stack(self.ctx, 'service_list_outputs_stack', tmpl)

        self.patchobject(self.eng, '_get_stack')
        self.patchobject(parser.Stack, 'load', return_value=stack)

        output = self.eng.show_output(self.ctx, mock.ANY, 'test')
        self.assertEqual({'output_key': 'test', 'output_value': 'first',
                          'description': 'sec'},
                         output)

        # Ensure that stack raised NotFound error with incorrect key.
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.eng.show_output,
                               self.ctx, mock.ANY, 'bunny')
        self.assertEqual(exception.NotFound, ex.exc_info[0])
        self.assertEqual('Specified output key bunny not found.',
                         str(ex.exc_info[1]))

    def test_stack_show_output_error(self):
        t = template_format.parse(tools.wp_template)
        t['outputs'] = {'test': {'value': 'first', 'description': 'sec'}}
        tmpl = templatem.Template(t)
        stack = parser.Stack(self.ctx, 'service_list_outputs_stack', tmpl)

        self.patchobject(self.eng, '_get_stack')
        self.patchobject(parser.Stack, 'load', return_value=stack)
        self.patchobject(
            stack.outputs['test'], 'get_value',
            side_effect=[exception.EntityNotFound(entity='one', name='name')])

        output = self.eng.show_output(self.ctx, mock.ANY, 'test')
        self.assertEqual(
            {'output_key': 'test',
             'output_error': "The one (name) could not be found.",
             'description': 'sec',
             'output_value': None},
            output)

    def test_stack_list_all_empty(self):
        sl = self.eng.list_stacks(self.ctx)

        self.assertEqual(0, len(sl))

    def test_stack_describe_all_empty(self):
        sl = self.eng.show_stack(self.ctx, None, resolve_outputs=True)

        self.assertEqual(0, len(sl))

    def test_get_template(self):
        # Setup
        t = template_format.parse(tools.wp_template)
        tmpl = templatem.Template(t)
        stack = parser.Stack(self.ctx, 'get_env_stack', tmpl)

        mock_get_stack = self.patchobject(self.eng, '_get_stack')
        mock_get_stack.return_value = mock.MagicMock()
        mock_get_stack.return_value.raw_template.template = t
        self.patchobject(parser.Stack, 'load', return_value=stack)

        # Test
        found = self.eng.get_template(self.ctx, stack.identifier())

        # Verify
        self.assertEqual(t, found)

    def test_get_template_no_template(self):
        # Setup
        exc = exception.EntityNotFound(entity='stack', name='missing')
        self.patchobject(self.eng, '_get_stack', side_effect=exc)

        # Test
        self.assertRaises(dispatcher.ExpectedException,
                          self.eng.get_template,
                          self.ctx,
                          'missing')

    def _preview_stack(self, environment_files=None):
        res._register_class('GenericResource1', generic_rsrc.GenericResource)
        res._register_class('GenericResource2', generic_rsrc.GenericResource)

        args = {}
        params = {}
        files = None
        stack_name = 'SampleStack'
        tpl = {'HeatTemplateFormatVersion': '2012-12-12',
               'Description': 'Lorem ipsum.',
               'Resources': {
                   'SampleResource1': {'Type': 'GenericResource1'},
                   'SampleResource2': {'Type': 'GenericResource2'}}}

        return self.eng.preview_stack(self.ctx, stack_name, tpl,
                                      params, files, args,
                                      environment_files=environment_files)

    def test_preview_stack_returns_a_stack(self):
        stack = self._preview_stack()
        expected_identity = {'path': '',
                             'stack_id': 'None',
                             'stack_name': 'SampleStack',
                             'tenant': 'stack_service_test_tenant'}
        self.assertEqual(expected_identity, stack['stack_identity'])
        self.assertEqual('SampleStack', stack['stack_name'])
        self.assertEqual('Lorem ipsum.', stack['description'])

    def test_preview_stack_returns_list_of_resources_in_stack(self):
        stack = self._preview_stack()
        self.assertIsInstance(stack['resources'], list)
        self.assertEqual(2, len(stack['resources']))

        resource_types = set(r['resource_type'] for r in stack['resources'])
        self.assertIn('GenericResource1', resource_types)
        self.assertIn('GenericResource2', resource_types)

        resource_names = set(r['resource_name'] for r in stack['resources'])
        self.assertIn('SampleResource1', resource_names)
        self.assertIn('SampleResource2', resource_names)

    def test_preview_stack_validates_new_stack(self):
        exc = exception.StackExists(stack_name='Validation Failed')
        self.eng._validate_new_stack = mock.Mock(side_effect=exc)
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self._preview_stack)
        self.assertEqual(exception.StackExists, ex.exc_info[0])

    @mock.patch.object(service.api, 'format_stack_preview', new=mock.Mock())
    @mock.patch.object(service.parser, 'Stack')
    def test_preview_stack_checks_stack_validity(self, mock_parser):
        self.patchobject(policy.ResourceEnforcer, 'enforce_stack')
        exc = exception.StackValidationFailed(message='Validation Failed')
        mock_parsed_stack = mock.Mock()
        mock_parsed_stack.validate.side_effect = exc
        mock_parser.return_value = mock_parsed_stack
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self._preview_stack)
        self.assertEqual(exception.StackValidationFailed, ex.exc_info[0])

    @mock.patch.object(env_util, 'merge_environments')
    def test_preview_environment_files(self, mock_merge):
        # Setup
        environment_files = ['env_1']

        # Test
        self._preview_stack(environment_files=environment_files)

        # Verify
        mock_merge.assert_called_once_with(environment_files, None, {}, {})

    @mock.patch.object(stack_object.Stack, 'get_by_name')
    def test_validate_new_stack_checks_existing_stack(self, mock_stack_get):
        mock_stack_get.return_value = 'existing_db_stack'
        tmpl = templatem.Template(
            {'HeatTemplateFormatVersion': '2012-12-12'})
        self.assertRaises(exception.StackExists, self.eng._validate_new_stack,
                          self.ctx, 'test_existing_stack', tmpl)

    @mock.patch.object(stack_object.Stack, 'count_all')
    def test_validate_new_stack_checks_stack_limit(self, mock_db_count):
        cfg.CONF.set_override('max_stacks_per_tenant', 99)
        mock_db_count.return_value = 99
        template = templatem.Template(
            {'HeatTemplateFormatVersion': '2012-12-12'})
        self.assertRaises(exception.RequestLimitExceeded,
                          self.eng._validate_new_stack,
                          self.ctx, 'test_existing_stack', template)

    def test_validate_new_stack_checks_incorrect_keywords_in_resource(self):
        template = {'heat_template_version': '2013-05-23',
                    'resources': {
                        'Res': {'Type': 'GenericResource1'}}}
        parsed_template = templatem.Template(template)
        ex = self.assertRaises(exception.StackValidationFailed,
                               self.eng._validate_new_stack,
                               self.ctx, 'test_existing_stack',
                               parsed_template)
        msg = ('"Type" is not a valid keyword '
               'inside a resource definition')

        self.assertEqual(msg, str(ex))

    def test_validate_new_stack_checks_incorrect_sections(self):
        template = {'heat_template_version': '2013-05-23',
                    'unknown_section': {
                        'Res': {'Type': 'GenericResource1'}}}
        parsed_template = templatem.Template(template)
        ex = self.assertRaises(exception.StackValidationFailed,
                               self.eng._validate_new_stack,
                               self.ctx, 'test_existing_stack',
                               parsed_template)
        msg = 'The template section is invalid: unknown_section'
        self.assertEqual(msg, str(ex))

    def test_validate_new_stack_checks_resource_limit(self):
        cfg.CONF.set_override('max_resources_per_stack', 5)
        template = {'HeatTemplateFormatVersion': '2012-12-12',
                    'Resources': {
                        'Res1': {'Type': 'GenericResource1'},
                        'Res2': {'Type': 'GenericResource1'},
                        'Res3': {'Type': 'GenericResource1'},
                        'Res4': {'Type': 'GenericResource1'},
                        'Res5': {'Type': 'GenericResource1'},
                        'Res6': {'Type': 'GenericResource1'}}}
        parsed_template = templatem.Template(template)
        self.assertRaises(exception.RequestLimitExceeded,
                          self.eng._validate_new_stack,
                          self.ctx, 'test_existing_stack', parsed_template)

    def test_validate_new_stack_handle_assertion_error(self):
        tmpl = mock.MagicMock()
        expected_message = 'Expected assertion error'
        tmpl.validate.side_effect = AssertionError(expected_message)
        exc = self.assertRaises(AssertionError, self.eng._validate_new_stack,
                                self.ctx, 'stack_name', tmpl)
        self.assertEqual(expected_message, str(exc))

    @mock.patch('heat.engine.service.ThreadGroupManager',
                return_value=mock.Mock())
    @mock.patch.object(stack_object.Stack, 'get_all')
    @mock.patch.object(stack_object.Stack, 'get_by_id')
    @mock.patch('heat.engine.stack_lock.StackLock',
                return_value=mock.Mock())
    @mock.patch.object(parser.Stack, 'load')
    @mock.patch.object(context, 'get_admin_context')
    def test_engine_reset_stack_status(
            self,
            mock_admin_context,
            mock_stack_load,
            mock_stacklock,
            mock_get_by_id,
            mock_get_all,
            mock_thread):
        mock_admin_context.return_value = self.ctx

        db_stack = mock.MagicMock()
        db_stack.id = 'foo'
        db_stack.status = 'IN_PROGRESS'
        db_stack.status_reason = None

        unlocked_stack = mock.MagicMock()
        unlocked_stack.id = 'bar'
        unlocked_stack.status = 'IN_PROGRESS'
        unlocked_stack.status_reason = None

        unlocked_stack_failed = mock.MagicMock()
        unlocked_stack_failed.id = 'bar'
        unlocked_stack_failed.status = 'FAILED'
        unlocked_stack_failed.status_reason = 'because'

        mock_get_all.return_value = [db_stack, unlocked_stack]
        mock_get_by_id.side_effect = [db_stack, unlocked_stack_failed]

        fake_stack = mock.MagicMock()
        fake_stack.action = 'CREATE'
        fake_stack.id = 'foo'
        fake_stack.status = 'IN_PROGRESS'

        mock_stack_load.return_value = fake_stack

        lock1 = mock.MagicMock()
        lock1.get_engine_id.return_value = 'old-engine'
        lock1.acquire.return_value = None
        lock2 = mock.MagicMock()
        lock2.acquire.return_value = None
        mock_stacklock.side_effect = [lock1, lock2]

        self.eng.thread_group_mgr = mock_thread

        self.eng.reset_stack_status()

        mock_admin_context.assert_called()
        filters = {
            'status': parser.Stack.IN_PROGRESS,
            'convergence': False
        }
        mock_get_all.assert_called_once_with(self.ctx,
                                             filters=filters,
                                             show_nested=True)
        mock_get_by_id.assert_has_calls([
            mock.call(self.ctx, 'foo'),
            mock.call(self.ctx, 'bar'),
        ])
        mock_stack_load.assert_called_once_with(self.ctx,
                                                stack=db_stack)
        self.assertTrue(lock2.release.called)
        reason = ('Engine went down during stack %s' % fake_stack.action)
        mock_thread.start_with_acquired_lock.assert_called_once_with(
            fake_stack, lock1,
            fake_stack.reset_stack_and_resources_in_progress, reason
        )

    def test_parse_adopt_stack_data_without_parameters(self):
        cfg.CONF.set_override('enable_stack_adopt', True)
        template = {"heat_template_version": "2015-04-30",
                    "resources": {
                        "myres": {
                            "type": "OS::Cinder::Volume",
                            "properties": {
                                "name": "volname",
                                "size": "1"
                            }
                        }
                    }}

        # Assert no KeyError exception raised like before, when trying to
        # get parameters from adopt stack data which doesn't have it.
        args = {"adopt_stack_data": '''{}'''}
        self.eng._parse_template_and_validate_stack(
            self.ctx, 'stack_name', template, {}, {},
            None, None, args)

        args = {"adopt_stack_data": '''{
            "environment": {}
        }'''}
        self.eng._parse_template_and_validate_stack(
            self.ctx, 'stack_name', template, {}, {},
            None, None, args)

    def test_parse_adopt_stack_data_with_parameters(self):
        cfg.CONF.set_override('enable_stack_adopt', True)
        template = {"heat_template_version": "2015-04-30",
                    "parameters": {
                        "volsize": {"type": "number"}
                    },
                    "resources": {
                        "myres": {
                            "type": "OS::Cinder::Volume",
                            "properties": {
                                "name": "volname",
                                "size": {"get_param": "volsize"}
                            }
                        }
                    }}

        args = {"adopt_stack_data": '''{
            "environment": {
                "parameters": {
                    "volsize": 1
                }
            }}'''}
        stack = self.eng._parse_template_and_validate_stack(
            self.ctx, 'stack_name', template, {}, {},
            None, None, args)
        self.assertEqual(1, stack.parameters['volsize'])

    @mock.patch('heat.engine.service.ThreadGroupManager',
                return_value=mock.Mock())
    @mock.patch.object(stack_object.Stack, 'get_by_id')
    @mock.patch.object(parser.Stack, 'load')
    def test_stack_cancel_update_convergence_with_no_rollback(
            self, mock_load, mock_get_by_id, mock_tg):
        stk = mock.MagicMock()
        stk.id = 1
        stk.UPDATE = 'UPDATE'
        stk.IN_PROGRESS = 'IN_PROGRESS'
        stk.state = ('UPDATE', 'IN_PROGRESS')
        stk.status = stk.IN_PROGRESS
        stk.action = stk.UPDATE
        stk.convergence = True
        mock_load.return_value = stk
        self.patchobject(self.eng, '_get_stack')
        self.eng.thread_group_mgr.start = mock.MagicMock()
        with mock.patch.object(self.eng, 'worker_service') as mock_ws:
            mock_ws.stop_traversal = mock.Mock()
            # with rollback as false
            self.eng.stack_cancel_update(self.ctx, 1,
                                         cancel_with_rollback=False)
            self.assertTrue(self.eng.thread_group_mgr.start.called)
            call_args, _ = self.eng.thread_group_mgr.start.call_args
            # test ID of stack
            self.assertEqual(call_args[0], 1)
            # ensure stop_traversal should be called with stack
            self.assertEqual(call_args[1].func, mock_ws.stop_traversal)
            self.assertEqual(call_args[1].args[0], stk)

    @mock.patch('heat.engine.service.ThreadGroupManager',
                return_value=mock.Mock())
    @mock.patch.object(stack_object.Stack, 'get_by_id')
    @mock.patch.object(parser.Stack, 'load')
    def test_stack_cancel_update_convergence_with_rollback(
            self, mock_load, mock_get_by_id, mock_tg):
        stk = mock.MagicMock()
        stk.id = 1
        stk.UPDATE = 'UPDATE'
        stk.IN_PROGRESS = 'IN_PROGRESS'
        stk.state = ('UPDATE', 'IN_PROGRESS')
        stk.status = stk.IN_PROGRESS
        stk.action = stk.UPDATE
        stk.convergence = True
        stk.rollback = mock.MagicMock(return_value=None)
        mock_load.return_value = stk
        self.patchobject(self.eng, '_get_stack')
        self.eng.thread_group_mgr.start = mock.MagicMock()
        # with rollback as true
        self.eng.stack_cancel_update(self.ctx, 1,
                                     cancel_with_rollback=True)
        self.eng.thread_group_mgr.start.assert_called_once_with(
            1, stk.rollback)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import routes
import webob

from heat.api.cfn.v1 import signal
from heat.api.cfn.v1 import stacks
from heat.common import wsgi


class API(wsgi.Router):

    """WSGI router for Heat CloudFormation v1 API requests."""

    _actions = {
        'list': 'ListStacks',
        'create': 'CreateStack',
        'describe': 'DescribeStacks',
        'delete': 'DeleteStack',
        'update': 'UpdateStack',
        'cancel_update': 'CancelUpdateStack',
        'events_list': 'DescribeStackEvents',
        'validate_template': 'ValidateTemplate',
        'get_template': 'GetTemplate',
        'estimate_template_cost': 'EstimateTemplateCost',
        'describe_stack_resource': 'DescribeStackResource',
        'describe_stack_resources': 'DescribeStackResources',
        'list_stack_resources': 'ListStackResources',
    }

    def __init__(self, conf, **local_conf):
        self.conf = conf
        mapper = routes.Mapper()

        stacks_resource = stacks.create_resource(conf)

        mapper.resource("stack", "stacks", controller=stacks_resource,
                        collection={'detail': 'GET'})

        def conditions(action):
            api_action = self._actions[action]

            def action_match(environ, result):
                req = webob.Request(environ)
                env_action = req.params.get("Action")
                return env_action == api_action

            return {'function': action_match}

        for action in self._actions:
            mapper.connect("/", controller=stacks_resource, action=action,
                           conditions=conditions(action))

        mapper.connect("/", controller=stacks_resource, action="index")

        # Add controller which handles signals on resources like:
        # waitconditions and alarms.
        # This is not part of the main CFN API spec, hence handle it
        # separately via a different path
        signal_controller = signal.create_resource(conf)
        mapper.connect('/waitcondition/{arn:.*}',
                       controller=signal_controller,
                       action='update_waitcondition',
                       conditions=dict(method=['PUT']))
        mapper.connect('/signal/{arn:.*}',
                       controller=signal_controller,
                       action='signal',
                       conditions=dict(method=['POST']))

        super(API, self).__init__(mapper)



# -*- coding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib.metadata as importlib_metadata
from unittest import mock

from heat.common import pluginutils
from heat.tests import common


class TestPluginUtil(common.HeatTestCase):

    def test_log_fail_msg(self):
        ep = importlib_metadata.EntryPoint(
            name=None, group=None,
            value='package.module:attr [extra1, extra2]')

        exc = Exception('Something went wrong')
        pluginutils.log_fail_msg(mock.Mock(), ep, exc)
        self.assertIn("Something went wrong", self.LOG.output)
